"""Test code for a single node in the ICOtronic system

The code below contains shared code for:

- SHA/STH
- SMH
- STU
"""

# -- Imports ------------------------------------------------------------------

from asyncio import get_running_loop, sleep
from datetime import date, datetime
from pathlib import Path
from typing import Any
from unittest import IsolatedAsyncioTestCase

from dynaconf.utils.boxing import DynaBox
from netaddr import EUI
from semantic_version import Version

from icotronic.can.connection import Connection
from icotronic.can.node.sth import STH
from icotronic.can.status import State
from icotronic.cmdline.commander import Commander
from icotronic.config import settings
from icotronic.can.node.eeprom.status import EEPROMStatus
from icotronic.report.report import Report
from icotronic.test.unit.extended_test_result import ExtendedTestResult
from icotronic import __version__

# -- Classes ------------------------------------------------------------------


class TestAttribute:
    """Test attribute that should be set to a value that is not ``None`` once

    Args:

        value:
            The value of the test attribute

        required_type:
            Stores type information about value

        unit:
            Optional unit for test attribute value

        pdf:
            Stores if the test attribute should be added to the test report or
            not

    Examples:

        Create an “empty” test attribute with type int

        >>> attribute = TestAttribute(int)
        >>> print(attribute.value)
        None

        Set a test attribute value later using the correct type

        >>> attribute.value = 1
        >>> attribute.value = 2

        Using an incorrect type will fail

        >>> attribute.value = "Something" # doctest:+ELLIPSIS
        Traceback (most recent call last):
           ...
        ValueError: The value “Something” is not ... type “<class 'int'>”
        >>> attribute.value
        2

        Initialize a test attribute with a certain value

        >>> TestAttribute(int, 5)
        5

    """

    def __init__(
        self,
        required_type: Any,
        value: Any = None,
        unit: str = "",
        pdf: bool = True,
    ) -> None:

        self.required_type = required_type
        self.value = value
        self.unit = unit
        self.pdf = pdf

    def __repr__(self) -> str:
        """Get a textual representation of the test attribute

        Returns:

            The textual representation of the test attribute

        Examples:

            Get string representation of an attribute with unit

            >>> TestAttribute(float, 11.0, "%")
            11.0 %

        """

        return f"{self.value}" + (f" {self.unit}" if self.unit else "")

    @property
    def value(self) -> Any:
        """Return the current value of the test attribute"""

        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        """Set the value of the test attribute

        Args:

            value:
                The new value of the test attribute

        Raises:

            ValueError:
                If the value is not ``None`` or does not have the required type

        """
        if value is None:
            self._value = None
            return

        if not isinstance(value, self.required_type):
            raise ValueError(
                f"The value “{value}” is not an instance of the required type "
                f"“{self.required_type}”"
            )

        self._value = value


# Use inner test class so we do not execute test methods of base class in
# addition to the test method of the super class.
# Source: https://stackoverflow.com/questions/1323455


# pylint: disable=too-few-public-methods
class BaseTestCases:
    """Collection of base test classes"""

    class TestNode(IsolatedAsyncioTestCase):
        """This class contains shared test code for STH and STU

        You are not supposed to use this class directly. Instead use it as base
        class for your test class.

        Every subclass of this class must set the attribute ``node`` to an
        object of the correct class (of ``can.node``).

        Please note, that this class only connects to the STU. If you also want
        to connect to a sensor node, please overwrite the method ``_connect``.

        To add additional test attributes shown in the standard output and
        optionally the PDF, add them to the **class** variable ``attributes``
        of this class or a subclass.

        The various ``_test`` methods in this class can be used to run certain
        tests for a node as part of a test method (i.e. a method that starts
        with the string ``test``).

        """

        attributes = {
            "Status": TestAttribute(str),
            "Name": TestAttribute(str),
            "Serial Number": TestAttribute(str, pdf=False),
            "GTIN": TestAttribute(int, pdf=False),
            "Bluetooth Address": TestAttribute(EUI),
            "Production Date": TestAttribute(date, pdf=False),
            "Hardware Version": TestAttribute(Version),
            "Batch Number": TestAttribute(int, pdf=False),
            "Firmware Version": TestAttribute(Version),
            "Release Name": TestAttribute(str, pdf=False),
            "EEPROM Status": TestAttribute(EEPROMStatus, pdf=False),
            "OEM Data": TestAttribute(str, pdf=False),
            "Operating Time": TestAttribute(int, pdf=False, unit="s"),
            "Power Off Cycles": TestAttribute(int, pdf=False),
            "Power On Cycles": TestAttribute(int, pdf=False),
            "Product Name": TestAttribute(str, pdf=False),
            "Under Voltage Counter": TestAttribute(int, pdf=False),
            "Watchdog Reset Counter": TestAttribute(int, pdf=False),
        }

        @classmethod
        def setUpClass(cls):
            """Set up data for whole test"""

            # Add a basic PDF report
            # Subclasses should overwrite this attribute, if you want to change
            # the default arguments of the report class
            cls.report = Report()

            # We store attributes related to the connection, such as MAC
            # address only once. To do that we set `read_attributes` to true
            # after the test class gathered the relevant data.
            cls.read_attributes = False

        @classmethod
        def tearDownClass(cls):
            """Print attributes of tested STH after all test cases"""

            cls.__output_general_data()
            cls.__output_node_data()
            cls.report.build()

        @classmethod
        def __output_general_data(cls):
            """Print general information and add it to PDF report"""

            now = datetime.now()

            attributes = {
                "ICOtronic Version": TestAttribute(
                    str, str(__version__), pdf=True
                ),
                "Date": TestAttribute(str, now.strftime("%Y-%m-%d"), pdf=True),
                "Time": TestAttribute(str, now.strftime("%H:%M:%S"), pdf=True),
                "Operator": TestAttribute(
                    str, settings.operator.name, pdf=True
                ),
            }

            cls.__output_data(attributes, node_data=False)

        @classmethod
        def __output_node_data(cls):
            """Print node information and add it to PDF report"""

            attributes = {
                description: (
                    TestAttribute(
                        str,
                        str(attribute.value),
                        unit=attribute.unit,
                        pdf=attribute.pdf,
                    )
                    if isinstance(
                        attribute.value, (date, EEPROMStatus, EUI, Version)
                    )
                    else attribute
                )
                for description, attribute in cls.attributes.items()
                if attribute.value is not None
            }

            cls.__output_data(attributes)

        @classmethod
        def __output_data(
            cls, attributes: dict[str, TestAttribute], node_data=True
        ) -> None:
            """Output data to standard output and PDF report

            Args:

                attributes:
                    The test attributes that should be printed

                node_data:
                    Specifies if this method outputs node specific or general
                    data

            """

            # Only output something, if there is at least one attribute
            if not attributes:
                return

            max_length_description = max(
                (len(description) for description in attributes)
            )
            max_length_value = max((
                len(str(attribute.value)) for attribute in attributes.values()
            ))

            # Print attributes to standard output
            print("\n")
            header = "Attributes" if node_data else "General"
            print(header)
            print("—" * len(header))

            for description, attribute in attributes.items():
                print(
                    f"{description:{max_length_description}} "
                    + f"{attribute.value:>{max_length_value}}"
                )

            # Add attributes to PDF
            attributes_pdf = {
                description: attribute
                for description, attribute in attributes.items()
                if attribute.pdf
            }

            assert hasattr(cls, "report")
            for description, attribute in attributes_pdf.items():
                cls.report.add_attribute(
                    description, attribute.value, node_data
                )

        async def asyncSetUp(self):
            """Set up hardware before a single test case"""

            # Disable debug mode (set by IsolatedAsyncioTestCase) to improve
            # runtime of code: https://github.com/python/cpython/issues/82789
            get_running_loop().set_debug(False)

            # All tests methods that contain the text `disconnected` do not
            # initialize a connection
            if self._testMethodName.find("disconnected") >= 0:
                return

            await self._connect()

            cls = type(self)
            # Only read node specific data once, even if we run multiple tests
            if not cls.read_attributes:
                cls.attributes["Bluetooth Address"].value = (
                    await self.node.get_mac_address()
                )
                cls.attributes["Firmware Version"].value = (
                    await self.node.get_firmware_version()
                )
                cls.attributes["Release Name"].value = (
                    await self.node.get_firmware_release_name()
                )
                cls.read_attributes = True

        async def asyncTearDown(self):
            """Clean up after single test case"""

            # All tests methods that contain the text `disconnected` do not
            # initialize a Bluetooth connection
            if self._testMethodName.find("disconnected") >= 0:
                return

            await self._disconnect()

        def run(self, result=None):
            """Execute a single test

            We override this method to store data about the test outcome.

            Args:

                result:
                     The unit test result of the test

            """

            super().run(result)
            if (
                not result.last_test.status
                == ExtendedTestResult.TestInformation.Status.SKIPPED
            ):
                type(self).report.add_test_result(
                    self.shortDescription(), result
                )

        async def _connect(self):
            """Create a connection to the STU"""

            # pylint: disable=attribute-defined-outside-init
            self.connection = Connection()
            # pylint: disable=unnecessary-dunder-call
            self.node = await self.connection.__aenter__()
            # pylint: enable=unnecessary-dunder-call
            # pylint: enable=attribute-defined-outside-init
            await self.node.reset()
            # Wait for reset to take place
            await sleep(2)

        async def _disconnect(self):
            """Tear down connection to STU"""

            await self.connection.__aexit__(None, None, None)

        async def test_connection(self):
            """Check connection to node"""

            # The sensor nodes need a little more time to switch from the
            # “Startup” to the “Operating” state
            await sleep(1)

            # Just send a request for the state and check, if the result
            # matches our expectations.
            state = await self.node.get_state()

            expected_state = State(
                mode="Get", location="Application", state="Operating"
            )

            self.assertEqual(
                expected_state,
                state,
                f"Expected state “{expected_state}” does not match "
                f"received state “{state}”",
            )

        def _test_firmware_flash(
            self,
            chip: str,
            flash_location: str | Path,
        ):
            """Upload bootloader and application into node

            Args:

                chip:
                    The name of the chip that should be flashed

                flash_location:
                    The location of the flash image

            """

            image_filepath = Path(flash_location).expanduser().resolve()
            self.assertTrue(
                image_filepath.exists(),
                f"Firmware file {image_filepath} does not exist",
            )
            self.assertTrue(
                image_filepath.is_file(),
                f"Firmware file {image_filepath} is not a file",
            )

            Commander().upload_flash(chip, image_filepath)

        async def _test_eeprom_product_data(self, config: DynaBox) -> None:
            """Test if reading and writing the product data EEPROM page works

            Args:

                config
                    A configuration object that stores the various product data
                    attributes

            """

            cls = type(self)

            node = self.node

            # ========
            # = GTIN =
            # ========

            gtin = config.gtin
            await node.eeprom.write_gtin(gtin)
            cls.attributes["GTIN"].value = await node.eeprom.read_gtin()
            self.assertEqual(
                gtin,
                cls.attributes["GTIN"].value,
                f"Written GTIN “{gtin}” does not match read GTIN"
                f" “{cls.attributes['GTIN'].value}”",
            )

            # ====================
            # = Hardware Version =
            # ====================

            hardware_version = config.hardware_version
            await node.eeprom.write_hardware_version(hardware_version)
            cls.attributes["Hardware Version"].value = (
                await node.eeprom.read_hardware_version()
            )
            self.assertEqual(
                hardware_version,
                f"{cls.attributes['Hardware Version'].value}",
                f"Written hardware version “{hardware_version}” does not "
                + "match read hardware version"
                f" “{cls.attributes['Hardware Version'].value}”",
            )

            # ====================
            # = Firmware Version =
            # ====================

            await node.eeprom.write_firmware_version(
                cls.attributes["Firmware Version"].value
            )
            firmware_version = await node.eeprom.read_firmware_version()
            self.assertEqual(
                f"{cls.attributes['Firmware Version'].value}",
                f"{firmware_version}",
                "Written firmware version"
                f" “{cls.attributes['Firmware Version'].value}” does not "
                + f"match read firmware version “{firmware_version}”",
            )

            # ================
            # = Release Name =
            # ================

            # Originally we assumed that this value would be set by the
            # firmware itself. However, according to tests with an empty EEPROM
            # this is not the case.
            release_name = config.firmware.release_name
            await node.eeprom.write_release_name(release_name)
            cls.attributes["Release Name"].value = (
                await node.eeprom.read_release_name()
            )
            self.assertEqual(
                release_name,
                cls.attributes["Release Name"].value,
                f"Written firmware release name “{release_name}” does not "
                + "match read firmware release name"
                f" “{cls.attributes['Release Name'].value}”",
            )

            # =================
            # = Serial Number =
            # =================

            serial_number = config.serial_number
            await node.eeprom.write_serial_number(serial_number)
            cls.attributes["Serial Number"].value = (
                await node.eeprom.read_serial_number()
            )
            self.assertEqual(
                serial_number,
                cls.attributes["Serial Number"].value,
                f"Written serial number “{serial_number}” does not "
                + "match read serial number"
                f" “{cls.attributes['Serial Number'].value}”",
            )

            # ================
            # = Product Name =
            # ================

            product_name = config.product_name
            await node.eeprom.write_product_name(product_name)
            cls.attributes["Product Name"].value = (
                await node.eeprom.read_product_name()
            )
            self.assertEqual(
                product_name,
                cls.attributes["Product Name"].value,
                f"Written product name “{product_name}” does not "
                + "match read product name"
                f" “{cls.attributes['Product Name'].value}”",
            )

            # ============
            # = OEM Data =
            # ============

            oem_data = config.oem_data
            await node.eeprom.write_oem_data(oem_data)
            oem_data_list = await node.eeprom.read_oem_data()
            self.assertListEqual(
                oem_data,
                oem_data_list,
                f"Written OEM data “{oem_data}” does not "
                + f"match read OEM data “{oem_data_list}”",
            )
            # We currently store the data in text format, to improve the
            # readability of null bytes in the shell. Please notice, that this
            # will not always work (depending on the binary data stored in
            # EEPROM region).
            cls.attributes["OEM Data"].value = "".join(
                map(chr, oem_data_list)
            ).replace("\x00", "")

        async def _test_eeprom_statistics(
            self, production_date: date, batch_number: int
        ) -> None:
            """Test if reading and writing the statistics EEPROM page works

            For this purpose this method writes (default) values into the
            EEPROM, reads them and then checks if the written and read values
            are equal.

            Args:

                production_date:
                    The production date of the node

                batch_number:
                    The batch number of the node

            """

            cls = type(self)
            node = self.node

            # =======================
            # = Power On/Off Cycles =
            # =======================

            power_on_cycles = 0
            await node.eeprom.write_power_on_cycles(power_on_cycles)
            cls.attributes["Power On Cycles"].value = (
                await node.eeprom.read_power_on_cycles()
            )
            self.assertEqual(
                power_on_cycles,
                cls.attributes["Power On Cycles"].value,
                f"Written power on cycle value “{power_on_cycles}” "
                + "does not match read power on cycle value "
                + f"“{cls.attributes['Power On Cycles'].value}”",
            )

            power_off_cycles = 0
            await node.eeprom.write_power_off_cycles(power_off_cycles)
            cls.attributes["Power Off Cycles"].value = (
                await node.eeprom.read_power_off_cycles()
            )
            self.assertEqual(
                power_off_cycles,
                cls.attributes["Power Off Cycles"].value,
                f"Written power off cycle value “{power_off_cycles}” "
                + "does not match read power off cycle value "
                + f"“{cls.attributes['Power Off Cycles'].value}”",
            )

            # ==================
            # = Operating Time =
            # ==================

            operating_time = 0
            await node.eeprom.write_operating_time(operating_time)
            cls.attributes["Operating Time"].value = (
                await node.eeprom.read_operating_time()
            )
            self.assertEqual(
                operating_time,
                cls.attributes["Operating Time"].value,
                f"Written operating time “{operating_time}” "
                + "does not match read operating time"
                " “{cls.attributes['Operating Time'].value}”",
            )

            # =========================
            # = Under Voltage Counter =
            # =========================

            under_voltage_counter = 0
            await node.eeprom.write_under_voltage_counter(
                under_voltage_counter
            )
            cls.attributes["Under Voltage Counter"].value = (
                await node.eeprom.read_under_voltage_counter()
            )
            self.assertEqual(
                under_voltage_counter,
                cls.attributes["Under Voltage Counter"].value,
                "Written under voltage counter value"
                f" “{under_voltage_counter}” "
                + "does not match read under voltage counter value "
                + f"“{cls.attributes['Under Voltage Counter'].value}”",
            )

            # ==========================
            # = Watchdog Reset Counter =
            # ==========================

            watchdog_reset_counter = 0
            await node.eeprom.write_watchdog_reset_counter(
                watchdog_reset_counter
            )
            cls.attributes["Watchdog Reset Counter"].value = (
                await node.eeprom.read_watchdog_reset_counter()
            )
            self.assertEqual(
                watchdog_reset_counter,
                cls.attributes["Watchdog Reset Counter"].value,
                "Written watchdog reset counter value"
                f" “{watchdog_reset_counter} does not match read watchdog"
                " reset counter value"
                f" “{cls.attributes['Watchdog Reset Counter'].value}”",
            )

            # ===================
            # = Production Date =
            # ===================

            await node.eeprom.write_production_date(production_date)
            cls.attributes["Production Date"].value = (
                await node.eeprom.read_production_date()
            )
            self.assertEqual(
                production_date,
                cls.attributes["Production Date"].value,
                f"Written production date “{production_date}” does not match "
                + "read production date"
                f" “{cls.attributes['Production Date'].value}”",
            )

            # ================
            # = Batch Number =
            # ================

            await node.eeprom.write_batch_number(batch_number)
            cls.attributes["Batch Number"].value = (
                await node.eeprom.read_batch_number()
            )
            self.assertEqual(
                batch_number,
                cls.attributes["Batch Number"].value,
                f"Written batch “{batch_number}” does not match "
                + "read batch number"
                f" “{cls.attributes['Batch Number'].value}”",
            )

        async def _test_eeprom_status(self) -> None:
            """Test if reading and writing the EEPROM status byte works"""

            cls = type(self)
            node = self.node

            # =================
            # = EEPROM Status =
            # =================

            await node.eeprom.write_status("Initialized")
            cls.attributes["EEPROM Status"].value = (
                await node.eeprom.read_status()
            )
            self.assertTrue(
                cls.attributes["EEPROM Status"].value.is_initialized(),
                "Setting EEPROM status to “Initialized” failed. "
                "EEPROM status byte currently stores the value "
                f"“{cls.attributes['EEPROM Status'].value}”",
            )

    class TestSensorNode(TestNode):
        """This class contains support code for sensor node (SMH & STH)

        You are not supposed to use this class directly, but instead use it as
        superclass for your test class. For more information, please take a
        look at the documentation of ``TestNode``.

        """

        @classmethod
        def setUpClass(cls):
            """Set up data for whole test"""

            super().setUpClass()

            cls.attributes["RSSI"] = TestAttribute(int, unit="dBm")

            cls.attributes["Serial Number"] = TestAttribute(str)

            cls.attributes["Ratio Noise Maximum"] = TestAttribute(
                float, unit="dB"
            )
            cls.attributes["Sleep Time 1"] = TestAttribute(
                int, unit="ms", pdf=False
            )
            cls.attributes["Advertisement Time 1"] = TestAttribute(
                int, unit="ms", pdf=False
            )
            cls.attributes["Sleep Time 2"] = TestAttribute(
                int, unit="ms", pdf=False
            )
            cls.attributes["Advertisement Time 2"] = TestAttribute(
                int, unit="ms", pdf=False
            )

        async def _connect_node(self, name: str) -> None:
            """Create a connection to the node with the specified name

            Args:

                name:
                    The (Bluetooth advertisement) name of the sensor node

            """

            await super()._connect()  # Connect to STU
            stu = self.node

            # pylint: disable=attribute-defined-outside-init
            self.sensor_node_connection = stu.connect_sensor_node(name, STH)
            # New node is sensor node
            # pylint: disable=unnecessary-dunder-call
            self.node = await self.sensor_node_connection.__aenter__()
            # pylint: enable=unnecessary-dunder-call
            self.stu = stu

            cls = type(self)
            cls.attributes["RSSI"].value = await self.node.get_rssi()

        async def _disconnect_node(self) -> None:
            """Disconnect from sensor node and STU"""

            await self.sensor_node_connection.__aexit__(None, None, None)
            await super()._disconnect()  # Disconnect from STU

        async def _test_name(self, name: str) -> str:
            """Check if writing and reading the name of a sensor node works

            Args:

                name:
                    The text that should be used as name for the sensor node

            Returns:

                Read back name

            """

            node = self.node
            await node.eeprom.write_name(name)
            read_name = await node.eeprom.read_name()

            self.assertEqual(
                name,
                read_name,
                f"Written name “{name}” does not match read name"
                f" “{read_name}”",
            )

            return read_name

        async def _test_eeprom_sleep_advertisement_times(self):
            """Test if reading and writing sleep/advertisement times works"""

            async def read_write_time(
                read_function,
                write_function,
                attribute,
                milliseconds,
            ):
                await write_function(milliseconds)
                milliseconds_read = round(await read_function())
                type(self).attributes[attribute].value = milliseconds_read
                self.assertEqual(
                    milliseconds_read,
                    milliseconds,
                    f"Value {milliseconds_read} ms of “{attribute}” does not "
                    f"match written value of {milliseconds} ms",
                )

            await read_write_time(
                read_function=self.node.eeprom.read_sleep_time_1,
                write_function=self.node.eeprom.write_sleep_time_1,
                attribute="Sleep Time 1",
                milliseconds=settings.sensor_node.bluetooth.sleep_time_1,
            )

            await read_write_time(
                read_function=self.node.eeprom.read_advertisement_time_1,
                write_function=self.node.eeprom.write_advertisement_time_1,
                attribute="Advertisement Time 1",
                milliseconds=(
                    settings.sensor_node.bluetooth.advertisement_time_1
                ),
            )

            await read_write_time(
                read_function=self.node.eeprom.read_sleep_time_2,
                write_function=self.node.eeprom.write_sleep_time_2,
                attribute="Sleep Time 2",
                milliseconds=settings.sensor_node.bluetooth.sleep_time_2,
            )

            await read_write_time(
                read_function=self.node.eeprom.read_advertisement_time_2,
                write_function=self.node.eeprom.write_advertisement_time_2,
                attribute="Advertisement Time 2",
                milliseconds=(
                    settings.sensor_node.bluetooth.advertisement_time_2
                ),
            )


# pylint: enable=too-few-public-methods

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Support for calculating dataloss statistics of ICOtronic CAN messages"""

# -- Imports ------------------------------------------------------------------

from collections.abc import Iterable

# -- Classes ------------------------------------------------------------------


class MessageStats:
    """Store message statistics

    Args:

        retrieved:
            The number of successfully retrieved messages

    """

    def __init__(self, retrieved: int = 0, lost: int = 0):

        self.retrieved = retrieved
        """Number of retrieved messages/streaming data elements"""

        self.lost = lost
        """Number of lost messages/streaming data elements"""

    def __repr__(self):
        """Get the textual representation of the message statistics

        Returns:

            A string representing the message statistics

        Examples:

            Get string representation of example data

            >>> MessageStats(retrieved=950, lost=50)
            Retrieved: 950, Lost: 50, Dataloss: 0.05

        """

        return ", ".join([
            f"Retrieved: {self.retrieved}",
            f"Lost: {self.lost}",
            f"Dataloss: {self.dataloss()}",
        ])

    def dataloss(self) -> float:
        """Get the amount of data loss

        Returns:

            The overall amount of dataloss as number between 0 (no data loss)
            and 1 (all data lost).

        Examples:

            Get the dataloss for some example data

            >>> MessageStats().dataloss()
            0
            >>> MessageStats(50, 50).dataloss()
            0.5
            >>> MessageStats(retrieved=960, lost=40).dataloss()
            0.04
            >>> MessageStats(retrieved=490, lost=10).dataloss()
            0.02
            >>> MessageStats(retrieved=0, lost=0).dataloss()
            0

        """

        overall = self.retrieved + self.lost

        return 0 if overall == 0 else self.lost / overall

    def reset(self) -> None:
        """Reset the amount of retrieved and lost messages to 0

        Examples:

            Reset stats for some example data

            >>> stats = MessageStats(10, 90)
            >>> stats.dataloss()
            0.9
            >>> stats.reset()
            >>> stats.dataloss()
            0

        """

        self.retrieved = 0
        self.lost = 0


# -- Functions ----------------------------------------------------------------


def calculate_dataloss_stats(counters: Iterable[int]) -> MessageStats:
    """Determine number of lost and received messages based on message counters

    Returns:

        Tuple containing the number of received and the number of lost
        messages

    Examples:

        Get data loss statistics for example counters

        >>> counters = [counter for counter in range(256)]
        >>> counters.extend([counter for counter in range(128, 256)])
        >>> stats = calculate_dataloss_stats(counters)
        >>> stats.retrieved
        384
        >>> stats.lost
        128

        >>> counters = [1, 1, 1, 9, 9, 9, 10, 10, 10]
        >>> calculate_dataloss_stats(counters)
        Retrieved: 3, Lost: 7, Dataloss: 0.7

    """

    iterator = iter(counters)

    try:
        last_counter = next(iterator)
    except StopIteration:
        return MessageStats()

    assert last_counter is not None
    lost_messages = 0
    retrieved_messages = 1

    for counter in iterator:
        if counter == last_counter:
            continue  # Skip data with same message counter

        retrieved_messages += 1

        lost_messages += (counter - last_counter) % 256 - 1

        last_counter = counter

    return MessageStats(retrieved=retrieved_messages, lost=lost_messages)



"""Communicate and control an ICOtronic node"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from semantic_version import Version

from icotronic.can.node.eeprom.node import NodeEEPROM
from icotronic.can.node.id import NodeId
from icotronic.can.node.spu import SPU
from icotronic.can.protocol.message import Message
from icotronic.can.status import State
from icotronic.test.misc import skip_hardware_tests_ci
from icotronic.utility.data import convert_bytes_to_text

pytestmark = skip_hardware_tests_ci()


# -- Classes ------------------------------------------------------------------


class Node:
    """Contains functionality shared by STU and sensor nodes

    Args:

        spu:
            The SPU object used to communicate with the node

        eeprom:
            The EEPROM class of the node

        id:
            The node identifier for the node

    """

    def __init__(
        self, spu: SPU, eeprom_class: type[NodeEEPROM], node_id: NodeId
    ) -> None:

        self.spu = spu
        self.id = node_id
        self.eeprom = eeprom_class(spu, node_id)

    # ==========
    # = System =
    # ==========

    async def reset(self) -> None:
        """Reset the node

        Examples:

            Import required library code

            >>> from asyncio import run, sleep
            >>> from icotronic.can.connection import Connection

            Reset the current STU

            >>> async def reset():
            ...     async with Connection() as stu:
            ...         await stu.reset()
            >>> run(reset())

            Reset a sensor node

            >>> async def reset():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             await sensor_node.reset()
            ...             # Wait some time for reset to take place
            ...             await sleep(1)
            >>> run(reset())

        """

        node = self.id
        message = Message(
            block="System",
            block_command="Reset",
            sender=self.spu.id,
            receiver=node,
            request=True,
        )
        await self.spu.request(
            message,
            description=f"reset node “{node}”",
            response_data=message.data,
            minimum_timeout=1,
        )

    # -----------------
    # - Get/Set State -
    # -----------------

    async def get_state(self) -> State:
        """Get the current state of the node

        Returns:

            The state of the node

        Examples:

            Import required library code

            >>> from asyncio import run, sleep
            >>> from icotronic.can.connection import Connection

            Get the state of the current STU

            >>> async def get_state():
            ...     async with Connection() as stu:
            ...         return await stu.get_state()
            >>> run(get_state())
            Get State, Location: Application, State: Operating

            Get state of sensor node

            >>> async def get_state():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             state = await sensor_node.get_state()
            ...             # Sensor node might be still in startup state
            ...             while state.state_name() == 'Startup':
            ...                 await sleep(1)
            ...                 state = await sensor_node.get_state()
            ...             return state
            >>> run(get_state())
            Get State, Location: Application, State: Operating

        """

        node = self.id
        message = Message(
            block="System",
            block_command="Get/Set State",
            sender=self.spu.id,
            receiver=node,
            request=True,
            data=[(State(mode="Get")).value],
        )

        response = await self.spu.request(
            message, description=f"get state of node “{node}”"
        )

        return State(response.data[0])

    # ================
    # = Product Data =
    # ================

    async def get_gtin(self) -> int:
        """Retrieve the GTIN (Global Trade Identification Number) of the node

        Returns:

            The Global Trade Identification Number

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the GTIN of STU 1

            >>> async def read_gtin():
            ...     async with Connection() as stu:
            ...         return await stu.get_gtin()
            >>> gtin = run(read_gtin())
            >>> isinstance(gtin, int)
            True

        """

        node = self.id
        response = await self.spu.request_product_data(
            node=node,
            description=f"read GTIN of node “{node}”",
            block_command="GTIN",
        )

        return int.from_bytes(response.data, byteorder="little")

    async def get_hardware_version(self) -> Version:
        """Retrieve the hardware version of a node

        Returns:

            The hardware version of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the hardware version of STU 1

            >>> async def read_hardware_version():
            ...     async with Connection() as stu:
            ...         return await stu.get_hardware_version()
            >>> hardware_version = run(read_hardware_version())
            >>> hardware_version.major
            1

        """

        node = self.id
        response = await self.spu.request_product_data(
            node=node,
            description=f"read hardware version of node “{node}”",
            block_command="Hardware Version",
        )

        major, minor, patch = response.data[-3:]
        return Version(major=major, minor=minor, patch=patch)

    async def get_firmware_version(self) -> Version:
        """Retrieve the firmware version of the node

        Returns:

            The firmware version of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the firmware version of STU 1

            >>> async def read_firmware_version():
            ...     async with Connection() as stu:
            ...         return await stu.get_firmware_version()
            >>> firmware_version = run(read_firmware_version())
            >>> firmware_version.major
            2

        """

        node = self.id
        response = await self.spu.request_product_data(
            node=node,
            description=f"read firmware version of node “{node}”",
            block_command="Firmware Version",
        )

        major, minor, patch = response.data[-3:]
        return Version(major=major, minor=minor, patch=patch)

    async def get_firmware_release_name(self) -> str:
        """Retrieve the firmware release name of a node

        Returns:

            The firmware release name of the specified node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the firmware release name of STU 1

            >>> async def read_release_name():
            ...     async with Connection() as stu:
            ...         return await stu.get_firmware_release_name()
            >>> run(read_release_name())
            'Valerie'

        """

        node = self.id
        response = await self.spu.request_product_data(
            node=node,
            description=f"read firmware release name of node “{node}”",
            block_command="Release Name",
        )

        release_name = convert_bytes_to_text(response.data, until_null=True)
        return release_name

    async def get_serial_number(self) -> str:
        """Retrieve the serial number of a node

        Returns:

            The serial number of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the serial number of STU 1

            >>> async def read_serial_number():
            ...     async with Connection() as stu:
            ...         return await stu.get_serial_number()
            >>> serial_number = run(read_serial_number())
            >>> isinstance(serial_number, str)
            True
            >>> 0 <= len(serial_number) <= 32
            True

        """

        async def get_serial_number_part(part: int) -> bytearray:
            """Retrieve a part of the serial number"""
            node = self.id
            response = await self.spu.request_product_data(
                node=node,
                description=(
                    f"read part {part} of the serial number of node “{node}”"
                ),
                block_command=f"Serial Number {part}",
            )
            return response.data

        serial_number_bytes = bytearray()
        for part in range(1, 5):
            serial_number_bytes.extend(await get_serial_number_part(part))

        return convert_bytes_to_text(serial_number_bytes)

    async def get_product_name(self) -> str:
        """Retrieve the product name of a node

        Returns:

            The product name of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the product name of STU 1

            >>> async def read_product_name():
            ...     async with Connection() as stu:
            ...         return await stu.get_product_name()
            >>> product_name = run(read_product_name())
            >>> isinstance(product_name, str)
            True
            >>> 0 <= len(product_name) <= 128
            True

        """

        async def get_product_name_part(part: int) -> bytearray:
            """Retrieve a part of the product name"""

            node = self.id
            response = await self.spu.request_product_data(
                node=node,
                description=(
                    f"read part {part} of the product name of node “{node}”"
                ),
                block_command=f"Product Name {part}",
            )
            return response.data

        product_name_bytes = bytearray()
        for part in range(1, 17):
            product_name_bytes.extend(await get_product_name_part(part))

        return convert_bytes_to_text(product_name_bytes)

    async def get_oem_data(self) -> bytearray:
        """Retrieve the OEM (free use) data

        Returns:

            The OEM data of the node

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read the OEM data of STU 1

            >>> async def read_oem_data():
            ...     async with Connection() as stu:
            ...         return await stu.get_oem_data()
            >>> oem_data = run(read_oem_data())
            >>> isinstance(oem_data, bytearray)
            True
            >>> len(oem_data)
            64

        """

        async def get_oem_part(part: int) -> bytearray:
            """Retrieve a part of the OEM data"""
            node = self.id
            response = await self.spu.request_product_data(
                node=node,
                description=(
                    f"read part {part} of the OEM data of node “{node}”"
                ),
                block_command=f"OEM Free Use {part}",
            )
            return response.data

        oem_data = bytearray()
        for part in range(1, 9):
            oem_data.extend(await get_oem_part(part))

        return oem_data


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import run_docstring_examples

    run_docstring_examples(
        Node.get_state,
        globals(),
        verbose=True,
    )



"""Data streaming support"""

# -- Exports ------------------------------------------------------------------

from icotronic.can.streaming.error import (
    StreamingError,
    StreamingTimeoutError,
    StreamingBufferError,
)
from icotronic.can.streaming.buffer import AsyncStreamBuffer
from icotronic.can.streaming.config import StreamingConfiguration
from icotronic.can.streaming.data import StreamingData
from icotronic.can.streaming.format import (
    StreamingFormat,
    StreamingFormatVoltage,
)



"""Check EEPROM of ICOtronic node"""

# -- Imports ------------------------------------------------------------------

from argparse import ArgumentParser, Namespace
from asyncio import run
from collections import Counter

from icotronic.can import Connection
from icotronic.cmdline.parse import byte_value, mac_address

# -- Function -----------------------------------------------------------------


def parse_arguments() -> Namespace:
    """Parse the arguments of the EEPROM checker command line tool

    Returns:

        A simple object storing the MAC address (attribute ``mac``) of an STH
        and an byte value that should be stored into the cells of the EEPROM
        (attribute ``value``)

    """

    parser = ArgumentParser(
        description="Check the integrity of STH EEPROM content"
    )
    parser.add_argument(
        "mac",
        help="MAC address of STH e.g. 08:6b:d7:01:de:81",
        type=mac_address,
    )
    parser.add_argument(
        "--value",
        help="byte value for EEPROM cells (default: %(default)s)",
        type=byte_value,
        default=10,
    )

    return parser.parse_args()


# -- Class --------------------------------------------------------------------


class EEPROMCheck:
    """Write and check the content of a sensor node EEPROM page

    Args:

        sensor_node:
            The sensor node where the EEPROM check should take place

        value:
            The value that the EEPROM checker should write into the EEPROM

    """

    def __init__(self, sensor_node, value):

        self.sensor_node = sensor_node
        self.eeprom_address = 1
        self.eeprom_length = 256
        self.eeprom_value = value

    async def write_eeprom(self):
        """Write a byte value into one page of the EEPROM"""

        print(f"Write value “{self.eeprom_value}” into EEPROM cells")
        await self.sensor_node.eeprom.write(
            address=1,
            offset=0,
            data=[self.eeprom_value for _ in range(self.eeprom_length)],
        )

    async def read_eeprom(self):
        """Read a page of the EEPROM

        Returns:

            A list of the byte values stored in the EEPROM page

        """

        return await self.sensor_node.eeprom.read(
            address=self.eeprom_address,
            offset=0,
            length=self.eeprom_length,
        )

    async def print_eeprom_incorrect(self):
        """Print a summary of the incorrect values in the EEPROM page"""

        changed = [
            byte
            for byte in await self.read_eeprom()
            if byte != self.eeprom_value
        ]
        incorrect = len(changed) / self.eeprom_length
        counter = Counter(changed)
        summary = ", ".join(
            f"{value} ({times} time{'' if times == 1 else 's'})"
            for value, times in sorted(
                counter.items(), key=lambda item: item[1], reverse=True
            )
        )
        print(f"{incorrect:.2%} incorrect{': ' if summary else ''}{summary}")

    async def print_eeprom(self):
        """Print the values stored in the EEPROM page"""

        page = await self.read_eeprom()
        bytes_per_line = 8
        for byte in range(0, self.eeprom_length - 1, bytes_per_line):
            print(f"{byte:3}: ", end="")
            byte_representation = " ".join(["{:3}"] * bytes_per_line).format(
                *page[byte : byte + bytes_per_line]
            )
            print(byte_representation)


# -- Functions ----------------------------------------------------------------


async def check_eeprom(arguments: Namespace):
    """Check EEPROM functionality

    Args:

        arguments:
            Command line arguments

    """

    times = 5

    async with Connection() as stu:
        async with stu.connect_sensor_node(arguments.mac) as sensor_node:
            print(f"Connected to node “{await sensor_node.get_name()}”")
            check = EEPROMCheck(sensor_node, arguments.value)
            await check.write_eeprom()
            await check.print_eeprom_incorrect()
            print()
        for counter in range(times):
            async with stu.connect_sensor_node(arguments.mac) as sensor_node:
                check = EEPROMCheck(sensor_node, arguments.value)
                await check.print_eeprom_incorrect()
                print()
                if counter >= times - 1:
                    await check.print_eeprom()


# -- Main ---------------------------------------------------------------------


def main():
    """Check EEPROM of node specified via command line argument"""
    run(check_eeprom(parse_arguments()))


if __name__ == "__main__":
    main()



"""Read and write EEPROM data of ICOtronic sensor nodes"""

# -- Imports ------------------------------------------------------------------

from icotronic.can.constants import ADVERTISEMENT_TIME_EEPROM_TO_MS
from icotronic.can.node.eeprom.node import NodeEEPROM
from icotronic.test.misc import skip_hardware_tests_ci

pytestmark = skip_hardware_tests_ci()


# -- Sensor -------------------------------------------------------------------


class SensorNodeEEPROM(NodeEEPROM):
    """Read and write EEPROM data of sensor nodes"""

    # ========================
    # = System Configuration =
    # ========================

    async def read_sleep_time_1(self) -> int:
        """Retrieve sleep time 1 from the EEPROM

        Returns:

            The current value of sleep time 1 in milliseconds

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read sleep time 1 of the sensor node with node id 0

            >>> async def read_sleep_time_1():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await sensor_node.eeprom.read_sleep_time_1()
            >>> sleep_time = run(read_sleep_time_1())
            >>> isinstance(sleep_time, int)
            True

        """

        return await self.read_int(address=0, offset=9, length=4)

    async def write_sleep_time_1(self, milliseconds: int) -> None:
        """Write the value of sleep time 1 to the EEPROM

        Args:

            milliseconds:
                The value for sleep time 1 in milliseconds

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read sleep time 1 of the sensor node with node id 0

            >>> async def write_read_sleep_time_1(milliseconds):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             await sensor_node.eeprom.write_sleep_time_1(
            ...                 milliseconds)
            ...             return await sensor_node.eeprom.read_sleep_time_1()
            >>> run(write_read_sleep_time_1(300_000))
            300000

        """

        await self.write_int(address=0, offset=9, value=milliseconds, length=4)

    async def read_advertisement_time_1(self) -> float:
        """Retrieve advertisement time 1 from the EEPROM

        Returns:

            The current value of advertisement time 1 in milliseconds

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read advertisement time 1 of of the sensor node with node id 0

            >>> async def read_advertisement_time_1():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await (
            ...                 sensor_node.eeprom.read_advertisement_time_1())
            >>> advertisement_time = run(read_advertisement_time_1())
            >>> isinstance(advertisement_time, float)
            True
            >>> advertisement_time > 0
            True

        """

        advertisement_time_eeprom = await self.read_int(
            address=0, offset=13, length=2
        )
        return advertisement_time_eeprom * ADVERTISEMENT_TIME_EEPROM_TO_MS

    async def write_advertisement_time_1(self, milliseconds: int):
        """Write the value of advertisement time 1 to the EEPROM

        Args:

            milliseconds:
                The value for advertisement time 1 in milliseconds

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read advertisement time 1 of sensor node with node id 0

            >>> async def write_read_advertisement_time_1(milliseconds):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             await (
            ...                 sensor_node.eeprom.write_advertisement_time_1(
            ...                     milliseconds))
            ...             return await (
            ...                 sensor_node.eeprom.read_advertisement_time_1())
            >>> run(write_read_advertisement_time_1(1250))
            1250.0

        """

        advertisement_time_eeprom = round(
            milliseconds / ADVERTISEMENT_TIME_EEPROM_TO_MS
        )

        await self.write_int(
            address=0,
            offset=13,
            value=advertisement_time_eeprom,
            length=2,
        )

    async def read_sleep_time_2(self) -> int:
        """Retrieve sleep time 2 from the EEPROM

        Returns:

            The current value of sleep time 2 in milliseconds

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read sleep time 2 of sensor node with node id 0

            >>> async def read_sleep_time_2():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await sensor_node.eeprom.read_sleep_time_2()
            >>> sleep_time = run(read_sleep_time_2())
            >>> isinstance(sleep_time, int)
            True

        """

        return await self.read_int(address=0, offset=15, length=4)

    async def write_sleep_time_2(self, milliseconds: int) -> None:
        """Write the value of sleep time 2 to the EEPROM

        Args:

            milliseconds:
                The value for sleep time 2 in milliseconds

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read sleep time 2 of sensor node with node id 0

            >>> async def write_read_sleep_time_2(milliseconds):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             await (sensor_node.eeprom.write_sleep_time_2(
            ...                    milliseconds))
            ...             return await sensor_node.eeprom.read_sleep_time_2()
            >>> run(write_read_sleep_time_2(259_200_000))
            259200000

        """

        await self.write_int(
            address=0, offset=15, value=milliseconds, length=4
        )

    async def read_advertisement_time_2(self) -> float:
        """Retrieve advertisement time 2 from the EEPROM

        Returns:

            The current value of advertisement time 2 in milliseconds

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Read advertisement time 2 of sensor node with node id 0

            >>> async def read_advertisement_time_2():
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             return await (
            ...                 sensor_node.eeprom.read_advertisement_time_2())
            >>> advertisement_time = run(read_advertisement_time_2())
            >>> isinstance(advertisement_time, float)
            True

        """

        advertisement_time_eeprom = await self.read_int(
            address=0, offset=19, length=2
        )

        return advertisement_time_eeprom * ADVERTISEMENT_TIME_EEPROM_TO_MS

    async def write_advertisement_time_2(self, milliseconds: int):
        """Write the value of advertisement time 2 to the EEPROM

        Args:

            milliseconds:
                The value for advertisement time 2 in milliseconds

        Examples:

            Import required library code

            >>> from asyncio import run
            >>> from icotronic.can.connection import Connection

            Write and read advertisement time 2 of sensor node with node id 0

            >>> async def write_read_advertisement_time_2(milliseconds):
            ...     async with Connection() as stu:
            ...         # We assume that at least one sensor node is available
            ...         async with stu.connect_sensor_node(0) as sensor_node:
            ...             await (
            ...                 sensor_node.eeprom.write_advertisement_time_2(
            ...                 milliseconds))
            ...             return await (
            ...                 sensor_node.eeprom.read_advertisement_time_2())
            >>> run(write_read_advertisement_time_2(2500))
            2500.0

        """

        advertisement_time_eeprom = round(
            milliseconds / ADVERTISEMENT_TIME_EEPROM_TO_MS
        )

        await self.write_int(
            address=0, offset=19, value=advertisement_time_eeprom, length=2
        )


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import run_docstring_examples

    run_docstring_examples(
        SensorNodeEEPROM.write_advertisement_time_2,
        globals(),
        verbose=True,
    )



"""CAN error handling support"""

# -- Classes ------------------------------------------------------------------


class UnsupportedFeatureException(Exception):
    """Indicate that a certain feature is not supported by a node"""


class CANConnectionError(Exception):
    """Exception for errors regarding the ICOtronic CAN connection"""


class CANInitError(CANConnectionError):
    """Exception for CAN initialization problems"""


class ErrorResponseError(CANConnectionError):
    """Exception for erroneous response messages"""


class NoResponseError(CANConnectionError):
    """Thrown if no response message for a request was received"""



"""Extensions to unit test code for our hardware tests"""

# -- Exports ------------------------------------------------------------------

from .extended_test_runner import ExtendedTestRunner
from .extended_test_result import ExtendedTestResult



"""Support code for Simplicity Commander command line tool

See: https://community.silabs.com/s/article/simplicity-commander

for more information
"""

# -- Imports ------------------------------------------------------------------

from os import environ, pathsep
from pathlib import Path
from platform import system
from re import compile as re_compile
from subprocess import CalledProcessError, run
from sys import byteorder

from icotronic.config import settings
from icotronic.test.misc import skip_hardware_tests_ci

pytestmark = skip_hardware_tests_ci()

# -- Classes ------------------------------------------------------------------


class CommanderException(Exception):
    """Describes problems regarding the execution of Simplicity Commander"""


class CommanderReturnCodeException(CommanderException):
    """A Simplicity Commander command did not return the success status code"""


class CommanderOutputMatchException(CommanderException):
    """The Simplicity Commander output did not contain an expected value"""


class Commander:
    """Wrapper for the Simplicity Commander commandline tool"""

    def __init__(self):

        self._add_path_to_environment()

        self.error_reasons = {
            "programmer not connected": (
                "Programming board is not connected to computer"
            ),
            "device not connected": (
                "Programming board is not connected to device"
            ),
        }

    def _add_path_to_environment(self) -> None:
        """Add path to Simplicity Commander (``commander``) to ``PATH``

        After calling this method you should be able to call ``commander``
        without its path prefix, if you installed it somewhere in the
        locations specified below ``COMMANDS`` → ``PATH`` in the configuration.

        Examples:

            Check that adding the commander path to the environment works

            >>> commander = Commander()

            >>> from subprocess import run
            >>> result = run("commander --version".split(),
            ...              capture_output=True)
            >>> result.returncode == 0
            True

        """

        path = settings.commands.path
        operating_system = system()
        paths = (
            path.linux
            if operating_system == "Linux"
            else path.mac if operating_system == "Darwin" else path.windows
        )

        environ["PATH"] += pathsep + pathsep.join(paths)

    def _run_command(
        self,
        command: list[str],
        description: str,
        possible_error_reasons: list[str] | None = None,
        regex_output: str | None = None,
    ) -> str:
        """Run a Simplicity Commander command

        Args:

            command:
                The Simplicity Commander subcommand including all necessary
                arguments

            description:
                A textual description of the purpose of the command
                e.g. “enable debug mode”

            possible_error_reasons:
                A list of dictionary keys that describe why the command might
                have failed

            regex_output:
                An optional regular expression that has to match part of the
                standard output of the command

        Raises:

            CommanderException

                - if the command returned unsuccessfully or

                - if the standard output did not match the optional regular
                  expression specified in ``regex_output``

        Returns:

            The standard output of the command

        """

        if possible_error_reasons:
            for reason in possible_error_reasons:
                if reason not in self.error_reasons:
                    raise ValueError(
                        f"“{reason}” is not a valid possible error reason"
                    )

        try:
            result = run(
                ["commander"] + command,
                capture_output=True,
                check=True,
                text=True,
            )
        except CalledProcessError as error:
            # Since Windows seems to return the exit code as unsigned number we
            # need to convert it first to the “real” signed number.
            returncode = (
                int.from_bytes(
                    error.returncode.to_bytes(4, byteorder),
                    byteorder,
                    signed=True,
                )
                if system() == "Windows"
                else error.returncode
            )
            error_message = (
                "Execution of Simplicity Commander command to "
                f"{description} failed with return code "
                f"“{returncode}”"
            )
            combined_output = (
                "\n".join((error.stdout, error.stderr))
                if error.stdout or error.stderr
                else ""
            )
            if combined_output:
                error_message += (
                    "\n\nSimplicity Commander output:\n\n"
                    f"{combined_output.rstrip()}"
                )

            if possible_error_reasons:
                error_reasons = "\n".join([
                    f"• {self.error_reasons[reason]}"
                    for reason in possible_error_reasons
                ])
                error_message += (
                    f"\n\nPossible error reasons:\n\n{error_reasons}"
                )

            raise CommanderReturnCodeException(error_message) from error

        if (
            regex_output is not None
            and re_compile(regex_output).search(result.stdout) is None
        ):
            error_message = (
                "Output of Simplicity Commander command to "
                f"{description}:\n{result.stdout}\n"
                "did not match the expected regular expression "
                f"“{regex_output}”"
            )
            raise CommanderOutputMatchException(error_message)

        return result.stdout

    def enable_debug_mode(self) -> None:
        """Enable debug mode for external device

        Examples:

            Import required library code

            >>> from icotronic.config import settings

            Enable debug mode of STH programming board

            >>> commander = Commander()
            >>> commander.enable_debug_mode()

        """

        error_reasons = ["programmer not connected"]
        self._run_command(
            command="adapter dbgmode OUT".split(),
            description="enable debug mode",
            possible_error_reasons=error_reasons,
            regex_output="Setting debug mode to OUT",
        )

    def unlock_device(self, chip: str) -> None:
        """Unlock device for debugging

        Warning:
            Calling this method will erase the flash of the device!

        Args:

            chip:

                The identifier of the chip on the PCB e.g. “BGM121A256V2”

        """

        self._run_command(
            command="device unlock".split() + ["-d", f"{chip}"],
            description="unlock device",
            possible_error_reasons=[
                "device not connected",
                "programmer not connected",
            ],
            regex_output="Chip successfully unlocked",
        )

    def upload_flash(self, chip: str, filepath: str | Path) -> None:
        """Upload code into the flash memory of the device

        Args:

            chip:
                The identifier of the chip on the PCB e.g. “BGM121A256V2”

            filepath:
                The filepath of the flash image

        Examples:

            Trying to upload a non existent file causes an error

            >>> commander = Commander()
            >>> # We assume the following file does not exist on your machine
            >>> filepath = "nothing here"
            >>> commander.upload_flash("BGM121A256V2",
            ...     filepath)  # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
               ...
            CommanderException: Firmware file “nothing here” does not exist

            Trying to upload a directory instead of a file results in an error

            >>> commander.upload_flash("BGM121A256V2",
            ...     Path.home()) # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
               ...
            CommanderException: “...” is not a file

        """

        firmware_filepath = Path(filepath)

        if not firmware_filepath.exists():
            raise CommanderException(
                f"Firmware file “{filepath}” does not exist"
            )

        if not firmware_filepath.is_file():
            raise CommanderException(f"“{filepath}” is not a file")

        # Set debug mode to out, to make sure we flash the STH (connected via
        # debug cable) and not another microcontroller connected to the
        # programmer board.
        self.enable_debug_mode()

        # Unlock device (triggers flash erase)
        self.unlock_device(chip)

        self._run_command(
            command=[
                "flash",
                f"{firmware_filepath}",
                "-d",
                f"{chip}",
                "--address",
                "0x0",
            ],
            description="upload firmware",
        )

    def read_power_usage(self, milliseconds: float = 1000) -> float:
        """Read the power usage of the connected hardware

        Args:

            milliseconds:
                The amount of time the power usage should be measured for

        Returns:

            The measured power usage in milliwatts

        Examples:

            Import required library code

            >>> from icotronic.config import settings

            Measure power usage of connected STH

            >>> commander = Commander()
            >>> commander.read_power_usage() > 0
            True

        """

        command = [
            "aem",
            "measure",
            "--windowlength",
            str(milliseconds),
        ]

        regex = r"Power\s*\[mW\]\s*:\s*(?P<milliwatts>\d+\.\d+)"
        try:
            output = self._run_command(
                command=command,
                description="read power usage",
                possible_error_reasons=["programmer not connected"],
                regex_output=regex,
            )
        except CommanderOutputMatchException as error:
            raise CommanderOutputMatchException(
                "Unable to extract power usage "
                "from Simplicity Commander output"
            ) from error

        pattern_match = re_compile(regex).search(output)
        assert pattern_match is not None
        milliwatts = pattern_match["milliwatts"]

        return float(milliwatts)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Data formats for storing streaming data"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from collections.abc import Callable

# -- Classes ------------------------------------------------------------------


class StreamingData:
    """Support for storing data of a streaming message

    Args:

        counter:
            The message counter value

        timestamp:
            The message timestamp

        values:
            The streaming values

    Examples:

        Create new streaming data

        >>> StreamingData(values=[1, 2, 3], counter=21, timestamp=1)
        [1, 2, 3]@1 #21

        Streaming data must store either two or three values

        >>> StreamingData(values=[1], counter=21, timestamp=1)
        Traceback (most recent call last):
        ...
        ValueError: Incorrect number of streaming values: 1 (instead of 2 or 3)

        >>> StreamingData(values=[1, 2, 3, 4], counter=21, timestamp=1
        ...              ) # doctest:+ELLIPSIS
        Traceback (most recent call last):
        ...
        ValueError: Incorrect number of ... values: 4 (instead of 2 or 3)

    """

    def __init__(
        self, counter: int, timestamp: float, values: list[float]
    ) -> None:

        if not 2 <= len(values) <= 3:
            raise ValueError(
                f"Incorrect number of streaming values: {len(values)} "
                "(instead of 2 or 3)"
            )

        self.counter = counter
        self.timestamp = timestamp
        self.values = values

    def apply(
        self,
        function: Callable[[float], float],
    ) -> StreamingData:
        """Apply a certain function to the streaming data

        Note:

            This function changes the stored values in the streaming data and
            (as convenience feature) also returns the modified streaming data
            itself. This is useful if you want to use the modified streaming
            as parameter in a function call, i.e. you can use something like
            ``function(stream_data.apply())``.

        Args:

            function:
                The function that should be applied to the streaming data

        Returns:

            The modified streaming data

        Examples:

            Add the constant 10 to some example streaming data

            >>> data = StreamingData(values=[1, 2, 3], counter=21, timestamp=1)
            >>> data.apply(lambda value: value + 10)
            [11, 12, 13]@1 #21
            >>> data.values
            [11, 12, 13]

        """

        updated_values = [function(value) for value in self.values]
        assert len(updated_values) == 2 or len(updated_values) == 3
        self.values = updated_values

        return self

    def __repr__(self):
        """Get the string representation of the stream data

        Examples:

            Get the string representation of some streaming data

            >>> StreamingData(values=[1, 2, 3], counter=21, timestamp=1)
            [1, 2, 3]@1 #21

        """

        return f"{self.values}@{self.timestamp} #{self.counter}"



"""ICOn command line tool

See: https://mytoolit.github.io/ICOtronic/#icon-cli-tool

for more information
"""

# -- Imports ------------------------------------------------------------------

from argparse import Namespace
from asyncio import run
from logging import basicConfig, getLogger
from sys import stderr
from tempfile import NamedTemporaryFile
from time import monotonic, perf_counter_ns, process_time_ns

from can.interfaces.pcan import PcanError
from tqdm import tqdm

from icotronic.can import Connection
from icotronic.can.adc import ADCConfiguration
from icotronic.can.error import CANConnectionError, UnsupportedFeatureException
from icotronic.can.node.sth import STH
from icotronic.can.sensor import SensorConfiguration
from icotronic.can.streaming import StreamingBufferError, StreamingTimeoutError
from icotronic.cmdline.parse import create_icon_parser
from icotronic.config import ConfigurationUtility, settings
from icotronic.measurement import Storage, StorageData
from icotronic.utility.performance import PerformanceMeasurement

# -- Functions ----------------------------------------------------------------


async def get_acceleration_sensor_range_in_g(sth: STH) -> float:
    """Read sensor range of acceleration sensor

    Args:

        sth:
            The STH object used to read the sensor range

    Returns:

        The sensor range of the acceleration sensor, or the default range of
        200 (± 100 g) sensor, if there was a problem reading the sensor range

    """

    sensor_range = 200

    try:
        sensor_range = await sth.get_acceleration_sensor_range_in_g()
        if sensor_range < 1:
            print(
                f"Warning: Sensor range “{sensor_range}” below 1 g — Using "
                "range 200 instead (± 100 g sensor)",
                file=stderr,
            )
            sensor_range = 200
    except ValueError:
        print(
            "Warning: Unable to determine sensor range from "
            "EEPROM value — Assuming ± 100 g sensor",
            file=stderr,
        )

    return sensor_range


def command_config() -> None:
    """Open configuration file"""

    ConfigurationUtility.open_user_config()


async def read_data(
    sth: STH,
    sensor_config: SensorConfiguration,
    storage: StorageData,
    measurement_time_s: float,
) -> PerformanceMeasurement:
    """Read some acceleration data from the given STH

    Args:

        sth:
            The STH from which data should be read

        sensor_config:
            The sensor configuration that should be used for reading data

        storage:
            The storage object that should be used to store the acceleration
            data

        measurement_time_s:
            The amount of time that should be used for reading data

    Returns:

        Information about the performance of the streaming code

    """

    streaming_config = sensor_config.streaming_configuration()
    logger = getLogger()
    logger.info("Streaming Configuration: %s", streaming_config)

    sample_rate = (await sth.get_adc_configuration()).sample_rate()
    progress = tqdm(
        total=int(sample_rate * measurement_time_s),
        desc="Read sensor data",
        unit=" values",
        leave=False,
        disable=None,
    )

    conversion_to_g = await sth.get_acceleration_conversion_function()
    values_per_message = streaming_config.data_length()

    performance_measurement = PerformanceMeasurement()
    try:
        async with sth.open_data_stream(streaming_config) as stream:
            performance_measurement.start()
            start_time = monotonic()
            async for data, _ in stream:
                storage.add_streaming_data(data.apply(conversion_to_g))
                progress.update(values_per_message)
                if monotonic() - start_time >= measurement_time_s:
                    break
            performance_measurement.stop()
    except PcanError as error:
        print(
            f"Unable to collect streaming data: {error}",
            file=stderr,
        )
    except KeyboardInterrupt:
        pass
    finally:
        progress.close()

    return performance_measurement


def print_dataloss_data(storage: StorageData) -> None:
    """Print information about data loss

    Args:

        storage:
            The data that should be analyzed for data loss

    """

    data_time_us = float(storage.measurement_time())
    data_time_s = data_time_us / 10**6
    dataloss = storage.dataloss()
    data_loss_status = (
        "🟢"
        if dataloss < 0.01
        else ("🟡" if dataloss < 0.05 else ("🟠" if dataloss < 0.1 else "🔴"))
    )

    dataloss_percent = storage.dataloss() * 100
    sample_rate_data = storage.sampling_frequency()
    sample_rate = storage["Sample_Rate"]
    print(
        "ADC:\n"
        f"  Sample Rate:      {sample_rate}\n"
        "Measurement:\n"
        f"  Sample Rate:      {sample_rate_data:.2f} Hz\n"
        f"  Data Loss:        {dataloss_percent:.2f} % "
        f"{data_loss_status}\n"
        f"  Measurement Time: {data_time_s:.2f} s"
    )


async def command_dataloss(arguments: Namespace) -> None:
    """Check data loss at different sample rates

    Args:

        arguments:
            The given command line arguments

    """

    identifier = arguments.identifier
    measurement_time_s = arguments.time

    logger = getLogger()

    async with Connection() as stu:
        logger.info("Connecting to “%s”", identifier)
        async with stu.connect_sensor_node(identifier, STH) as sth:
            assert isinstance(sth, STH)
            logger.info("Connected to “%s”", identifier)

            sensor_range = await get_acceleration_sensor_range_in_g(sth)
            sensor_config = SensorConfiguration(first=1)

            oversampling_rates = [2**exponent for exponent in range(6, 10)]
            for oversampling_rate in oversampling_rates:
                logger.info("Oversampling rate: %s", oversampling_rate)
                adc_config = ADCConfiguration(
                    prescaler=2,
                    acquisition_time=8,
                    oversampling_rate=oversampling_rate,
                )
                await sth.set_adc_configuration(**adc_config)
                logger.info("Sample rate: %s Hz", adc_config.sample_rate())

                with NamedTemporaryFile(
                    suffix=".hdf5", delete_on_close=False
                ) as temp:
                    logger.info("Temporary measurement file: %s", temp.name)
                    with Storage(
                        temp.name, sensor_config.streaming_configuration()
                    ) as storage:
                        storage.write_sensor_range(sensor_range)
                        storage.write_sample_rate(adc_config)

                        try:
                            performance = await read_data(
                                sth, sensor_config, storage, measurement_time_s
                            )
                        except StreamingBufferError as error:
                            print(f"⚠️ {error}", file=stderr)
                        finally:
                            print_dataloss_data(storage)

                        print("Performance:")
                        print(f"  {performance}")

                        print(
                            (
                                ""
                                if oversampling_rate == oversampling_rates[-1]
                                else "\n"
                            ),
                            end="",
                        )


async def command_list(
    arguments: Namespace,  # pylint: disable=unused-argument
) -> None:
    """Print a list of available sensor nodes

    Args:

        arguments:
            The given command line arguments

    """

    async with Connection() as stu:
        sensor_nodes = await stu.collect_sensor_nodes()

        for node in sensor_nodes:
            print(node)


async def command_measure(arguments: Namespace) -> None:
    """Open measurement stream and store data

    Args:

        arguments:
            The given command line arguments

    """

    identifier = arguments.identifier
    measurement_time_s = arguments.time

    async with Connection() as stu:
        async with stu.connect_sensor_node(identifier, STH) as sth:
            assert isinstance(sth, STH)

            adc_config = ADCConfiguration(
                reference_voltage=arguments.voltage_reference,
                prescaler=arguments.prescaler,
                acquisition_time=arguments.acquisition,
                oversampling_rate=arguments.oversampling,
            )
            await sth.set_adc_configuration(**adc_config)
            print(f"Sample Rate: {adc_config.sample_rate():.2f} Hz")

            user_sensor_config = SensorConfiguration(
                first=arguments.first_channel,
                second=arguments.second_channel,
                third=arguments.third_channel,
            )

            if user_sensor_config.requires_channel_configuration_support():
                try:
                    await sth.set_sensor_configuration(user_sensor_config)
                except UnsupportedFeatureException as exception:
                    raise UnsupportedFeatureException(
                        f"Sensor channel configuration “{user_sensor_config}”"
                        f" is not supported by the sensor node “{identifier}”"
                    ) from exception

            sensor_range = await get_acceleration_sensor_range_in_g(sth)
            filepath = settings.get_output_filepath()

            with Storage(
                filepath, user_sensor_config.streaming_configuration()
            ) as storage:
                storage.write_sensor_range(sensor_range)
                storage.write_sample_rate(adc_config)

                try:
                    await read_data(
                        sth, user_sensor_config, storage, measurement_time_s
                    )
                except KeyboardInterrupt:
                    pass
                finally:
                    print(f"Data Loss: {storage.dataloss() * 100} %")
                    print(f"Filepath: {filepath}")


async def command_rename(arguments: Namespace) -> None:
    """Rename a sensor node

    Args:

        arguments:
            The given command line arguments

    """

    identifier = arguments.identifier
    name = arguments.name

    async with Connection() as stu:
        async with stu.connect_sensor_node(identifier) as sensor_node:
            old_name = await sensor_node.get_name()
            mac_address = await sensor_node.get_mac_address()

            await sensor_node.set_name(name)
            name = await sensor_node.get_name()
            print(
                f"Renamed sensor node “{old_name}” with MAC "
                f"address “{mac_address}” to “{name}”"
            )


async def command_stu(arguments: Namespace) -> None:
    """Run specific commands regarding stationary transceiver unit

    Args:

        arguments:
            The given command line arguments

    """

    subcommand = arguments.stu_subcommand

    async with Connection() as stu:
        if subcommand == "mac":
            print(await stu.get_mac_address())
        elif subcommand == "reset":
            await stu.reset()
        else:
            raise ValueError(f"Unknown STU subcommand “{subcommand}”")


def main():
    """ICOtronic command line tool"""

    parser = create_icon_parser()
    arguments = parser.parse_args()
    try:
        if arguments.subcommand == "measure":
            SensorConfiguration(
                first=arguments.first_channel,
                second=arguments.second_channel,
                third=arguments.third_channel,
            ).check()
    except ValueError as error:
        parser.prog = f"{parser.prog} {arguments.subcommand}"
        parser.error(str(error))

    basicConfig(
        level=arguments.log.upper(),
        style="{",
        format="{asctime} {levelname:7} {message}",
    )

    logger = getLogger()
    logger.info("CLI Arguments: %s", arguments)

    if arguments.subcommand == "config":
        command_config()
    else:
        command_to_coroutine = {
            "dataloss": command_dataloss,
            "list": command_list,
            "measure": command_measure,
            "rename": command_rename,
            "stu": command_stu,
        }

        try:
            perf_start, cpu_start = perf_counter_ns(), process_time_ns()
            run(command_to_coroutine[arguments.subcommand](arguments))
            perf_end, cpu_end = perf_counter_ns(), process_time_ns()
            run_time_command = perf_end - perf_start
            cpu_time_command = cpu_end - cpu_start
            cpu_usage = cpu_time_command / run_time_command * 100
            logger.info(
                "Ran command “%s” in %.2f seconds (CPU time: %.2f seconds, "
                "CPU Usage: %.2f %%)",
                arguments.subcommand,
                run_time_command / 10**9,
                cpu_time_command / 10**9,
                cpu_usage,
            )
        except (
            CANConnectionError,
            StreamingBufferError,
            TimeoutError,
            UnsupportedFeatureException,
            ValueError,
        ) as error:
            print(error, file=stderr)
        except StreamingTimeoutError as error:
            print(f"Quitting Measurement: {error}")
        except KeyboardInterrupt:
            pass


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    main()



"""Support for storing additional information about test results"""

# -- Imports ------------------------------------------------------------------

from enum import auto, Enum
from unittest import TextTestResult

# -- Class --------------------------------------------------------------------


class ExtendedTestResult(TextTestResult):
    """Store data about the result of a test"""

    class TestInformation:
        """Store additional data of a test result

        We use this class to store test information in a PDF report.

        """

        class Status(Enum):
            """Store the status of a test"""

            SUCCESS = auto()
            FAILURE = auto()
            ERROR = auto()
            SKIPPED = auto()

        def __init__(self):

            self.status = type(self).Status.SUCCESS
            self.message = ""

        def set_error(self, message):
            """Set the status of the test to error

            Args:

                message:
                    Specifies the error message

            """

            self.status = type(self).Status.ERROR
            self.message = message

        def set_failure(self, message):
            """Set the status of the test to failure

            Args:

                message:
                    Specifies the failure message

            """

            self.status = type(self).Status.FAILURE
            self.message = message

        def set_success(self):
            """Set the status of the test to success"""

            self.status = type(self).Status.SUCCESS
            self.message = ""

        def set_skipped(self):
            """Set the status of the test to skipped"""

            self.status = type(self).Status.SKIPPED

        def error(self):
            """Check if there was an error

            Returns:

                ``True`` if there was an error, ``False`` otherwise

            """

            return self.status == type(self).Status.ERROR

        def failure(self):
            """Check if there test failed

            Returns:

                ``True`` if the test failed, ``False`` otherwise

            """

            return self.status == type(self).Status.FAILURE

    def __init__(self, *arguments, **keyword_arguments):

        super().__init__(*arguments, **keyword_arguments)

        self.last_test = ExtendedTestResult.TestInformation()

    def addFailure(self, test, err):
        """Add information about the latest failure

        Args:

            test:
                The test case that produced the failure

            err:
                A tuple of the form returned by ``sys.exc_info()``:
                ``(type, value, traceback)``

        """

        super().addFailure(test, err)

        # Store message for latest failure
        failure_message = str(err[1])
        # Only store custom message added to assertion, since it should be more
        # readable for a person. If there was no custom message, then the
        # object stores the auto-generated message.
        custom_failure_message = failure_message.rsplit(" : ", maxsplit=1)[-1]

        self.last_test.set_failure(custom_failure_message)

    def addError(self, test, err):
        """Add information about the latest error

        This should usually not happen unless there are problems with the
        connection or the syntax of the current code base.

        Args:

            test:
                The test case that produced the error

            err:
                A tuple of the form returned by ``sys.exc_info()``:
                ``(type, value, traceback)``

        """

        super().addError(test, err)

        self.last_test.set_error(err[1])

    def addSuccess(self, test):
        """Add information about latest successful test

        Args:

            test:
                The successful test

        """

        super().addSuccess(test)

        self.last_test.set_success()

    def addSkip(self, test, reason):
        """Add information about latest skipped test

        Args:

            test:
                The skipped test

            reason:
                The reason why the test was skipped

        """

        super().addSkip(test, reason)

        self.last_test.set_skipped()



"""Support code for sensors and sensor configuration"""

# -- Imports ------------------------------------------------------------------

from collections.abc import Iterable
from enum import auto, Enum
from statistics import mean
from typing import NamedTuple

# -- Classes ------------------------------------------------------------------


class SensorType(Enum):
    """Possible sensor types"""

    BROKEN = auto()
    """Broken sensor"""

    ACCELERATION = auto()
    """Acceleration sensor"""

    TEMPERATURE = auto()
    """Temperature sensor"""

    PIEZO = auto()
    """Piezoelectric sensor"""


class Sensor(NamedTuple):
    """Store information about a sensor"""

    type: SensorType
    """Sensor type"""

    mean: float
    """Mean value of raw ADC sensor data"""

    def __repr__(self) -> str:
        """Return a string representation of the sensor

        Returns:

            A string that describes the sensor

        Examples:

            Get the textual representation of some example sensor objects

            >>> Sensor(SensorType.BROKEN, mean=123)
            Broken Sensor (Mean: 123)

            >>> Sensor(SensorType.TEMPERATURE, mean=123)
            Temperature Sensor

        """

        representation = f"{self.type.name.capitalize()} Sensor"
        if self.type == SensorType.BROKEN:
            representation += f" (Mean: {self.mean})"

        return representation

    def works(self) -> bool:
        """Check if the sensor is working or not

        Returns:

            ``True`` if the sensor works, ``False`` otherwise

        """

        return not self.type == SensorType.BROKEN


# -- Functions ----------------------------------------------------------------


def guess_sensor(values: Iterable[int]) -> Sensor:
    """Guess the sensor type from raw 16 bit ADC values

    Args:

        values:
            Multiple raw 16 bit ADC measurement values

    Returns:

        An object representing the guessed sensor type

    Examples:

        Guess sensor types based on raw ADC values

        >>> guess_sensor([38024, 38000, 37950])
        Piezo Sensor

        >>> guess_sensor([32500, 32571, 32499])
        Acceleration Sensor

        >>> guess_sensor([10650, 10500, 10780])
        Temperature Sensor

        >>> guess_sensor([123, 10, 50])
        Broken Sensor (Mean: 61)

    """

    mean_raw = mean(values)
    half = 2**15

    tolerance_acceleration = 1000
    min_acceleration = half - tolerance_acceleration
    max_acceleration = half + tolerance_acceleration

    if 10000 <= mean_raw <= 11000:
        return Sensor(SensorType.TEMPERATURE, mean_raw)
    if min_acceleration <= mean_raw <= max_acceleration:
        return Sensor(SensorType.ACCELERATION, mean_raw)
    if 37500 <= mean_raw <= 38500:
        return Sensor(SensorType.PIEZO, mean_raw)

    return Sensor(SensorType.BROKEN, mean_raw)


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Test code for sensory tool holder (STH)/ assembly (SHA)

The difference between the SHA and the STH is that the SHA denotes only the
PCB while an STH also includes the tool holder that contains the SHA.
"""

# -- Imports ------------------------------------------------------------------

from itertools import chain, repeat
from time import monotonic
from unittest import main as unittest_main, skipIf

from semantic_version import Version

from icotronic.can.streaming import StreamingConfiguration
from icotronic.config import settings
from icotronic.measurement.acceleration import (
    convert_raw_to_g,
    ratio_noise_max,
)
from icotronic.measurement.constants import ADC_MAX_VALUE
from icotronic.report.report import Report
from icotronic.utility.naming import convert_mac_base64
from icotronic.test.unit import ExtendedTestRunner
from icotronic.test.production.node import BaseTestCases, TestAttribute

# -- Classes ------------------------------------------------------------------


class TestSTH(BaseTestCases.TestSensorNode):
    """This class contains tests for the Sensory Tool Holder (STH)"""

    @classmethod
    def setUpClass(cls):
        """Set up data for whole test"""

        super().setUpClass()

        cls.attributes["Status"] = TestAttribute(str, settings.sth.status)
        cls.attributes["Ratio Noise Maximum"] = TestAttribute(
            float,
            unit="dB",
            pdf=True,
        )
        cls.attributes["Holder Type"] = TestAttribute(
            str, settings.sth.holder_type
        )

        sensor_name = settings.sth.acceleration_sensor.sensor
        maximum_acceleration = (
            settings.acceleration_sensor().acceleration.maximum
        )
        cls.attributes["Acceleration Sensor"] = TestAttribute(
            str, f"±{maximum_acceleration // 2} g Sensor ({sensor_name})"
        )

        for axis in "xyz":
            cls.attributes[f"Acceleration Slope {axis.upper()}"] = (
                TestAttribute(float, pdf=False)
            )
            cls.attributes[f"Acceleration Offset {axis.upper()}"] = (
                TestAttribute(float, pdf=False)
            )

        cls.report = Report(node="STH")

        # Manual checks
        cls.report.add_checkbox_list(
            title="Metal Blank",
            boxes=[
                "Okay",
                "Cylindrical thread defect",
                "Dent",
                "Oil spillage",
                "Shavings",
                "Milling errors",
            ],
            text_fields=1,
        )

        cls.report.add_checkbox_list(
            title="PCB",
            boxes=["Optical inspection: no defects"],
            text_fields=1,
        )

        cls.report.add_checkbox_list(
            title="Before Resin Cast",
            boxes=[
                "Battery test successful",
                "Charge in charging station was successful",
            ],
            text_fields=2,
        )

        cls.report.add_checkbox_list(
            title="Final Checks",
            boxes=[
                "Resin cast contains no bubbles",
                "Resin cast hardened completely",
                "No resin residue outside of pocket",
                "Pocket is completely filled with resin",
                "No oil spillage in vacuum chamber",
                "Charge in charging station was successful",
            ],
            text_fields=2,
        )

    async def asyncSetUp(self):
        """Set up hardware before a single test case"""

        await super().asyncSetUp()

        # Sensor node is connected after set up function unless the test
        # does not initiate a Bluetooth connection, which is the case if
        # the test name (method name) contains the text “disconnected”.
        if self._testMethodName.find("disconnected") < 0:
            # Change reference voltage depending on acceleration sensor
            reference_voltage = (
                settings.acceleration_sensor().reference_voltage
            )

            await self.node.set_adc_configuration(
                prescaler=2,
                acquisition_time=8,
                oversampling_rate=64,
                reference_voltage=reference_voltage,
            )

    async def _connect(self):
        """Create a connection to the STH"""

        await super()._connect_node(settings.sth_name())

    async def _disconnect(self):
        """Tear down connection to STH"""

        await super()._disconnect_node()

    @skipIf(
        settings.sth.status == "Epoxied",
        f"Flash test skipped because of status “{settings.sth.status}”",
    )
    def test__firmware_flash_disconnected(self):
        """Upload bootloader and application into STH

        Note:

            The additional underscore in the method name that makes sure this
            test case is executed before all other test cases.

        Note:

            The text ``disconnected`` in the method name makes sure that the
            test framework does not initialize a connection.

        """

        try:
            hardware_version = Version(settings.sth.hardware_version)
        except (TypeError, ValueError) as error:
            raise ValueError(
                "Incorrect STH hardware version: "
                f"“{settings.sth.hardware_version}”"
            ) from error

        if not 1 <= hardware_version.major <= 2:
            raise ValueError(
                f"STH hardware version “{hardware_version}” "
                "is currently not supported"
            )

        chip = (
            "BGM113A256V2" if hardware_version.major <= 1 else "BGM123A256V2"
        )

        self._test_firmware_flash(
            flash_location=settings.sth.firmware.location.flash,
            chip=chip,
        )

    async def test_battery_voltage(self):
        """Test voltage of STH power source"""

        supply_voltage = await self.node.get_supply_voltage()
        expected_voltage = settings.sth.battery_voltage.average
        tolerance_voltage = settings.sth.battery_voltage.tolerance
        expected_minimum_voltage = expected_voltage - tolerance_voltage
        expected_maximum_voltage = expected_voltage + tolerance_voltage

        self.assertGreaterEqual(
            supply_voltage,
            expected_minimum_voltage,
            f"STH supply voltage of {supply_voltage:.3f} V is lower "
            "than expected minimum voltage of "
            f"{expected_minimum_voltage:.3f} V",
        )
        self.assertLessEqual(
            supply_voltage,
            expected_maximum_voltage,
            f"STH supply voltage of {supply_voltage:.3f} V is "
            "greater than expected maximum voltage of "
            f"{expected_minimum_voltage:.3f} V",
        )

    async def test_acceleration_single_value(self):
        """Test stationary acceleration value"""

        sensor = settings.acceleration_sensor()
        stream_data = await self.node.get_streaming_data_single()
        acceleration = convert_raw_to_g(
            stream_data.values[0], sensor.acceleration.maximum
        )

        # We expect a stationary acceleration between -g₀ and g₀
        # (g₀ = 9.807 m/s²)
        expected_acceleration = 0
        tolerance_acceleration = sensor.acceleration.tolerance
        expected_minimum_acceleration = (
            expected_acceleration - tolerance_acceleration
        )
        expected_maximum_acceleration = (
            expected_acceleration + tolerance_acceleration
        )

        self.assertGreaterEqual(
            acceleration,
            expected_minimum_acceleration,
            f"Measured acceleration {acceleration:.3f} g is lower "
            "than expected minimum acceleration "
            f"{expected_minimum_acceleration:.3f} g",
        )
        self.assertLessEqual(
            acceleration,
            expected_maximum_acceleration,
            f"Measured acceleration {acceleration:.3f} g is greater "
            "than expected maximum acceleration "
            f"{expected_maximum_acceleration:.3f} g",
        )

    async def test_acceleration_noise(self):
        """Test ratio of noise to maximal possible measurement value"""

        async def read_streaming_data():
            """Read streaming data of first channel"""
            stream_data = []
            seconds = 4
            async with self.node.open_data_stream(
                StreamingConfiguration(first=True)
            ) as stream:
                end_time = monotonic() + seconds
                async for data, _ in stream:
                    stream_data.extend(data.values)
                    if monotonic() > end_time:
                        break

            return stream_data

        acceleration = await read_streaming_data()

        ratio_noise_maximum = ratio_noise_max(acceleration)
        sensor = settings.acceleration_sensor()
        maximum_ratio_allowed = sensor.acceleration.ratio_noise_to_max_value

        self.assertLessEqual(
            ratio_noise_maximum,
            maximum_ratio_allowed,
            (
                "The ratio noise to possible maximum measured value of "
                f"{ratio_noise_maximum} dB "  # pylint: disable=no-member
                "is higher than the maximum allowed level of "
                f" {maximum_ratio_allowed} dB"
            ),
        )
        cls = type(self)
        cls.attributes["Ratio Noise Maximum"].value = round(
            ratio_noise_maximum, 2
        )

    async def test_acceleration_self_test(self):
        """Execute self test of accelerometer"""

        async def read_voltages(dimension, reference_voltage) -> list[int]:
            """Read acceleration voltages in millivolts"""

            before = await self.node.get_acceleration_voltage(
                dimension, reference_voltage
            )

            await self.node.activate_acceleration_self_test(dimension)
            between = await self.node.get_acceleration_voltage(
                dimension, reference_voltage
            )

            await self.node.deactivate_acceleration_self_test(dimension)
            after = await self.node.get_acceleration_voltage(
                dimension, reference_voltage
            )

            return [round(value * 1000) for value in (before, between, after)]

        sensor = settings.acceleration_sensor()

        (
            voltage_before_test,
            voltage_at_test,
            voltage_after_test,
        ) = await read_voltages(
            sensor.self_test.dimension, sensor.reference_voltage
        )

        voltage_diff = voltage_at_test - voltage_before_test
        voltage_diff_abs = abs(voltage_diff)
        voltage_diff_before_after = abs(
            voltage_before_test - voltage_after_test
        )

        # - Voltage difference can be both positive or negative
        # - Voltage before and after the self test should be roughly the same

        voltage_diff_expected = sensor.self_test.voltage.difference
        voltage_diff_tolerance = sensor.self_test.voltage.tolerance

        voltage_diff_minimum = voltage_diff_expected - voltage_diff_tolerance
        voltage_diff_maximum = voltage_diff_expected + voltage_diff_tolerance

        self.assertLessEqual(
            voltage_diff_before_after,
            voltage_diff_tolerance,
            "Measured voltage difference between voltage before and after "
            f"test {voltage_diff_before_after:.0f} mV is larger than "
            f"tolerance of {voltage_diff_tolerance:.0f} mV",
        )

        possible_failure_reason = (
            "\n\nPossible Reason:\n\n• Acceleration sensor config value "
            f"“{settings.sth.acceleration_sensor.sensor}” is incorrect"
        )

        self.assertGreaterEqual(
            voltage_diff_abs,
            voltage_diff_minimum,
            f"Measured voltage difference of {voltage_diff_abs:.0f} mV is "
            "lower than expected minimum voltage difference of "
            f"{voltage_diff_minimum:.0f} mV{possible_failure_reason}",
        )
        self.assertLessEqual(
            voltage_diff_abs,
            voltage_diff_maximum,
            f"Measured voltage difference of {voltage_diff_abs:.0f} mV is "
            "greater than expected minimum voltage difference of "
            f"{voltage_diff_maximum:.0f} mV{possible_failure_reason}",
        )

    # pylint: disable=too-many-locals

    async def test_eeprom(self):
        """Test if reading and writing the EEPROM works"""

        cls = type(self)

        # ========
        # = Name =
        # ========

        name = (
            settings.sth.serial_number
            if settings.sth.status == "Epoxied"
            else convert_mac_base64(cls.attributes["Bluetooth Address"].value)
        )
        cls.attributes["Name"].value = await self._test_name(name)

        # =========================
        # = Sleep & Advertisement =
        # =========================

        await self._test_eeprom_sleep_advertisement_times()

        # ================
        # = Product Data =
        # ================

        await self._test_eeprom_product_data(settings.sth)

        # ==============
        # = Statistics =
        # ==============

        await self._test_eeprom_statistics(
            settings.sth.production_date,
            settings.sth.batch_number,
        )

        # ================
        # = Acceleration =
        # ================

        sensor = settings.acceleration_sensor()
        acceleration_max = sensor.acceleration.maximum
        adc_max = ADC_MAX_VALUE
        acceleration_slope = acceleration_max / adc_max
        acceleration_offset = -(acceleration_max / 2)

        # pylint: disable=too-many-arguments, too-many-positional-arguments

        async def write_read_check(
            attribute, write_routine, value, read_routine, axis, name
        ):
            await write_routine(value)
            read_value = await read_routine()
            self.assertAlmostEqual(
                value,
                read_value,
                msg=(
                    f"Written {axis} acceleration {name} "
                    f"“{acceleration_slope:.5f}” does not match read "
                    f"{axis} acceleration {name} "
                    f"“{read_value:.5f}”"
                ),
            )
            # Round value for nicer output formatting
            cls.attributes[attribute].value = round(read_value, 5)

        # pylint: enable=too-many-arguments, too-many-positional-arguments

        class_variables = (
            "Acceleration Slope X",
            "Acceleration Slope Y",
            "Acceleration Slope Z",
            "Acceleration Offset X",
            "Acceleration Offset Y",
            "Acceleration Offset Z",
        )
        write_routines = (
            self.node.eeprom.write_x_axis_acceleration_slope,
            self.node.eeprom.write_y_axis_acceleration_slope,
            self.node.eeprom.write_z_axis_acceleration_slope,
            self.node.eeprom.write_x_axis_acceleration_offset,
            self.node.eeprom.write_y_axis_acceleration_offset,
            self.node.eeprom.write_z_axis_acceleration_offset,
        )
        read_routines = (
            self.node.eeprom.read_x_axis_acceleration_slope,
            self.node.eeprom.read_y_axis_acceleration_slope,
            self.node.eeprom.read_z_axis_acceleration_slope,
            self.node.eeprom.read_x_axis_acceleration_offset,
            self.node.eeprom.read_y_axis_acceleration_offset,
            self.node.eeprom.read_z_axis_acceleration_offset,
        )
        names = chain(*(repeat("slope", 3), repeat("offset", 3)))
        values = chain(*(
            repeat(acceleration_slope, 3),
            repeat(acceleration_offset, 3),
        ))
        axes = list("xyz") * 2

        for (
            class_variable,
            write_routine,
            value,
            read_routine,
            axis,
            name,
        ) in zip(
            class_variables,
            write_routines,
            values,
            read_routines,
            axes,
            names,
        ):
            await write_read_check(
                class_variable,
                write_routine,
                value,
                read_routine,
                axis,
                name,
            )

        # =================
        # = EEPROM Status =
        # =================

        await self._test_eeprom_status()

        # =========
        # = Reset =
        # =========

        # We reset the STH and STU to make sure
        # - the name change takes place and we can connect to the STH
        #   using the new name
        # - the STH also takes the other changed EEPROM values (such as
        #   the changed advertisement times) into account.
        await self.node.reset()
        await self.stu.reset()

        try:
            async with self.stu.connect_sensor_node(
                cls.attributes["Name"].value
            ):
                pass  # Reconnected to STH
        except TimeoutError:
            self.fail(
                "Unable to reconnect to STH using updated name "
                f"“{cls.attributes['Name'].value}”"
            )

    # pylint: enable=too-many-locals


def main():
    """Run production test for Sensory Tool Holder (STH)"""

    unittest_main(
        testRunner=ExtendedTestRunner, module="icotronic.test.production.sth"
    )


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    main()



"""Support for streaming (measurement) data in the ICOtronic system"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from asyncio import Queue, wait_for
from collections.abc import AsyncIterator
from time import time

from can import Listener, Message

from icotronic.can.protocol.identifier import Identifier
from icotronic.can.dataloss import MessageStats
from icotronic.can.streaming.error import (
    StreamingBufferError,
    StreamingTimeoutError,
)
from icotronic.can.streaming.data import StreamingData

# -- Classes ------------------------------------------------------------------


class AsyncStreamBuffer(Listener):
    """Buffer for streaming data

    Args:

        timeout:
            The amount of seconds between two consecutive messages, before
            a ``StreamingTimeoutError`` will be raised

        max_buffer_size:
            Maximum amount of buffered messages kept by the stream buffer.
            If this amount is exceeded, then this listener will raise a
            ``StreamingBufferError``. A large buffer indicates that the
            application is not able to keep up with the current rate of
            retrieved messages and therefore the probability of losing
            messages is quite high.

    """

    def __init__(
        self,
        timeout: float,
        max_buffer_size: int,
    ) -> None:

        # Expected identifier of received streaming messages
        self.identifier = Identifier(
            block="Streaming",
            block_command="Data",
            sender="STH 1",
            receiver="SPU 1",
            request=False,
        )
        self.queue: Queue[tuple[StreamingData, int]] = Queue()
        self.timeout = timeout
        self.last_counter = -1
        self.max_buffer_size = max_buffer_size
        self.stats = MessageStats()
        self.timestamp_offset: float | None = None

    def __aiter__(self) -> AsyncIterator[tuple[StreamingData, int]]:
        """Retrieve iterator for collected data

        Returns:

            An iterator over the received streaming data including the number
            of lost messages

        """

        return self

    async def __anext__(self) -> tuple[StreamingData, int]:
        """Retrieve next stream data object in collected data

        Returns:

            A tuple containing:

            - the data of the streaming message and
            - the number of lost streaming messages right before the returned
              streaming message

        """

        if self.queue.qsize() > self.max_buffer_size:
            raise StreamingBufferError(
                f"Maximum buffer size of {self.max_buffer_size} messages "
                "exceeded"
            )

        try:
            return await wait_for(self.queue.get(), self.timeout)
        except TimeoutError as error:
            raise StreamingTimeoutError(
                f"No data received for at least {self.timeout} seconds"
            ) from error

    def on_message_received(self, msg: Message) -> None:
        """Handle received messages

        Args:

            msg:
                The received CAN message

        """

        # Ignore messages with wrong id and “Stop Stream” messages
        if msg.arbitration_id != self.identifier.value or len(msg.data) <= 1:
            return

        receive_time = time()

        # Calculate timestamp offset for first received message
        if self.timestamp_offset is None:
            self.timestamp_offset = receive_time - msg.timestamp

        data = msg.data
        counter = data[1]
        timestamp = msg.timestamp + self.timestamp_offset
        data_bytes = (
            data[start : start + 2] for start in range(2, len(data) - 1, 2)
        )

        values: list[float] = [
            int.from_bytes(word, byteorder="little") for word in data_bytes
        ]
        assert len(values) == 2 or len(values) == 3

        streaming_data = StreamingData(
            timestamp=timestamp,
            counter=counter,
            values=values,
        )

        # Calculate amount of lost messages
        if self.last_counter < 0:
            self.last_counter = (counter - 1) % 256
        last_counter = self.last_counter
        lost_messages = (counter - last_counter) % 256 - 1
        self.last_counter = counter
        self.stats.lost += lost_messages
        self.stats.retrieved += 1

        self.queue.put_nowait((streaming_data, lost_messages))

    def on_error(self, exc: Exception) -> None:
        """This method is called to handle any exception in the receive thread.

        Args:

            exc:
                The exception causing the thread to stop

        """

        raise NotImplementedError()

    def stop(self) -> None:
        """Stop handling new messages"""

    def reset_stats(self) -> None:
        """Reset the message statistics

        This method resets the amount of lost an retrieved messages used in
        the calculation of the method ``dataloss``. Using this method can be
        useful, if you want to calculate the amount of data loss since a
        specific starting point.

        """

        self.stats.reset()

    def dataloss(self) -> float:
        """Calculate the overall amount of data loss

        Returns:

            The overall amount of data loss as number between 0 (no data loss)
            and 1 (all data lost).

        """

        return self.stats.dataloss()


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Protocol and CAN constants"""

ADVERTISEMENT_TIME_EEPROM_TO_MS = 0.625
"""Stores the conversion rate for the EEPROM advertisement times

See also:

- https://mytoolit.github.io/Documentation/#page-system-configuration

"""

SENSOR_NODE_NUMBER_SELF_ADDRESSING = 0xFF
"""Sensor Node Number for self addressing mode"""



"""Support for measuring data"""

# -- Exports ------------------------------------------------------------------

from .acceleration import convert_raw_to_g, ratio_noise_max
from .data import ChannelData, Conversion, DataPoint, MeasurementData
from .storage import Storage, StorageData



"""Type checking code"""


def check_list(
    data: object, required_length: int, argument_name="data"
) -> None:
    """Check if the given object is a list with the given length

    Args:

        data:
            The object that should be checked for list type

        required_length:
            The required minimum length of the object, if it is a list

        argument_name:
            The variable name of the given object

    Raises:

        ValueError:
            if the given object is not a list or is too long

    """

    if not isinstance(data, list):
        raise ValueError(
            f"Unsupported object type for argument {argument_name}: "
            f"“{type(data)}”"
        )
    if len(data) < required_length:
        raise ValueError(
            f"Data length of {len(data)} is too "
            "small, at least length of "
            f"“{required_length}” required"
        )



"""Streaming configuration support"""

# -- Imports ------------------------------------------------------------------

from ctypes import c_uint8, LittleEndianStructure

# -- Classes ------------------------------------------------------------------

# pylint: disable=too-few-public-methods


class StreamingConfigBits(LittleEndianStructure):
    """Store enable/disabled channels of streaming configuration"""

    _fields_ = [
        ("first", c_uint8, 1),
        ("second", c_uint8, 1),
        ("third", c_uint8, 1),
    ]


# pylint: enable=too-few-public-methods


class StreamingConfiguration:
    """Streaming configuration

    Args:

        first:
            Specifies if the first channel is enabled or not

        second:
            Specifies if the second channel is enabled or not

        third:
            Specifies if the third channel is enabled or not

    Raises:

        ValueError:
            if none of the channels is active

    Examples:

        Create some example streaming configurations

        >>> config = StreamingConfiguration()
        >>> config = StreamingConfiguration(
        ...          first=False, second=True, third=True)

        Creating streaming configurations without active channels will fail

        >>> config = StreamingConfiguration(first=False)
        Traceback (most recent call last):
           ...
        ValueError: At least one channel needs to be active

        >>> config = StreamingConfiguration(
        ...     first=False, second=False, third=False)
        Traceback (most recent call last):
           ...
        ValueError: At least one channel needs to be active

    """

    def __init__(
        self, first: bool = True, second: bool = False, third: bool = False
    ) -> None:

        if first + second + third <= 0:
            raise ValueError("At least one channel needs to be active")

        self.channels = StreamingConfigBits(
            first=first, second=second, third=third
        )

    def __repr__(self) -> str:
        """Return the string representation of the streaming configuration

        Examples:

            Get the textual representation of some example streaming configs

            >>> StreamingConfiguration()
            Channel 1 enabled, Channel 2 disabled, Channel 3 disabled

            >>> StreamingConfiguration(first=False, second=True, third=False)
            Channel 1 disabled, Channel 2 enabled, Channel 3 disabled

            >>> StreamingConfiguration(first=True, second=True, third=True)
            Channel 1 enabled, Channel 2 enabled, Channel 3 enabled

        """

        channels = self.channels

        return ", ".join([
            f"Channel {name} {'en' if status else 'dis'}abled"
            for name, status in enumerate(
                (channels.first, channels.second, channels.third), start=1
            )
        ])

    def enabled_channels(self) -> int:
        """Get the number of activated channels

        Returns:

            The number of enabled channels

        Examples:

            Get the number of enabled channels for example streaming configs

            >>> StreamingConfiguration(first=True).enabled_channels()
            1

            >>> StreamingConfiguration(first=False, second=True, third=False
            ...                       ).enabled_channels()
            1

            >>> StreamingConfiguration(first=True, second=True, third=True
            ...                       ).enabled_channels()
            3

        """

        channels = self.channels

        return channels.first + channels.second + channels.third

    def data_length(self) -> int:
        """Returns the streaming data length

        This will be either:

        - 2 (when 2 channels are active), or
        - 3 (when 1 or 3 channels are active)

        For more information, please take a look `here`_.

        .. _here: https://mytoolit.github.io/Documentation/#command-data

        Returns:

            The length of the streaming data resulting from this channel
            configuration

        Examples:

            Get the data length of example streaming configurations

            >>> StreamingConfiguration().data_length()
            3

            >>> StreamingConfiguration(
            ...     first=False, second=True, third=False).data_length()
            3

            >>> StreamingConfiguration(
            ...     first=True, second=True, third=True).data_length()
            3

            >>> StreamingConfiguration(
            ...     first=False, second=True, third=True).data_length()
            2

        """

        return 2 if self.enabled_channels() == 2 else 3

    def axes(self) -> list[str]:
        """Get the activated axes returned by this streaming configuration

        Returns:

            A list containing all activated axes in alphabetical order

        Examples:

            Get the activated axes for example streaming configurations

            >>> StreamingConfiguration(
            ...     first=False, second=True, third=True).axes()
            ['y', 'z']
            >>> StreamingConfiguration(
            ...     first=True, second=True, third=False).axes()
            ['x', 'y']

        """

        channels = self.channels
        return [
            axis
            for axis, status in zip(
                "xyz",
                (channels.first, channels.second, channels.third),
            )
            if status
        ]

    @property
    def first(self) -> bool:
        """Check the activation state of the first channel

        Returns:

        ``True``, if the first channel is enabled or ``False`` otherwise

        Examples:

            Check channel one activation status for example configs

            >>> StreamingConfiguration(first=True, second=False,
            ...                        third=False).first
            True
            >>> StreamingConfiguration(first=False, second=False,
            ...                        third=True).first
            False

        """

        return bool(self.channels.first)

    @property
    def second(self) -> bool:
        """Check the activation state of the second channel

        Returns:

            ``True``, if the second channel is enabled or ``False`` otherwise

        Examples:

            Check channel two activation status for example configs

            >>> StreamingConfiguration(
            ...     first=True, second=False, third=False).second
            False
            >>> StreamingConfiguration(
            ...     first=False, second=True, third=True).second
            True

        """

        return bool(self.channels.second)

    @property
    def third(self) -> bool:
        """Check the activation state of the third channel

        Returns:

        ``True``, if the third channel is enabled or ``False`` otherwise

        Examples:

            Check channel three activation status for example configs

            >>> StreamingConfiguration(
            ...     first=True, second=False, third=False).third
            False
            >>> StreamingConfiguration(
            ...     first=False, second=False, third=True).third
            True

        """

        return bool(self.channels.third)



"""Command Line Parsing support"""

# -- Imports ------------------------------------------------------------------

from argparse import ArgumentParser, ArgumentTypeError
from math import inf
from re import compile as re_compile

from netaddr import AddrFormatError, EUI

from icotronic.can.adc import ADCConfiguration

# -- Functions ----------------------------------------------------------------


def base64_mac_address(name):
    """Check if the given text represents a Base64 encoded MAC address

    Returns:

        The given text on success

    Raises:

        An argument type error in case the given value does not represent a MAC
        address

    Examples:

        Parse valid Base64 encoded MAC address

        >>> base64_mac_address("CGvXAd6B")
        'CGvXAd6B'

        Parse invalid Base64 encoded MAC address

        >>> base64_mac_address("CGvXAd") # doctest:+NORMALIZE_WHITESPACE
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “CGvXAd” is not a Base64 encoded MAC
                                    address

    """

    base64_regex = re_compile("[A-Za-z0-9/+]{8}$")
    if base64_regex.match(name):
        return name
    raise ArgumentTypeError(f"“{name}” is not a Base64 encoded MAC address")


def byte_value(value):
    """Check if the given string represents a byte value

    Returns:

        An integer representing the given value on success

    Raises:

        An argument type error in case the given value does not represent a
        (positive) byte value

    Examples:

        Parse valid byte values

        >>> byte_value("0xa")
        10
        >>> byte_value("137")
        137

        Parse incorrect byte value

        >>> byte_value("256")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “256” is not a valid byte value

    """

    try:
        number = int(value, base=0)
        if number < 0 or number > 255:
            raise ValueError()
        return number
    except ValueError as error:
        raise ArgumentTypeError(
            f"“{value}” is not a valid byte value"
        ) from error


def channel_number(value: str):
    """Check if the given string represents a valid channel number (0 – 255)

    Raises:

        An argument type error in case the given value does not represent a
        channel number

    Returns:

        An integer representing the given channel number on success

    Examples:

        Parse correct channel numbers

        >>> channel_number("1")
        1
        >>> channel_number("123")
        123
        >>> channel_number("0")
        0
        >>> channel_number("255")
        255

        Parsing an incorrect negative channel number will fail

        >>> channel_number("-1")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “-1” is not a valid channel number

        Parse an incorrectly large channel number will fail

        >>> channel_number("256")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “256” is not a valid channel number

    """

    try:
        number: int | None = None
        number = int(value)
        if number < 0 or number > 255:
            raise ValueError()
        return number
    except ValueError as error:
        raise ArgumentTypeError(
            f"“{number if number is not None else value}” is not a valid "
            "channel number"
        ) from error


def mac_address(address: str) -> EUI:
    """Check if the given text represents a MAC address

    Returns:

        The parsed MAC address

    Raises:

        An argument type error in case the given text does not store a MAC
        address of the form

        - ``xx:xx:xx:xx:xx:xx`` or
        - ``xx-xx-xx-xx-xx-xx``

        where `x` represents a hexadecimal number.

    Examples:

        Parse a correct MAC address

        >>> mac_address("08:6b:d7:01:de:81")
        EUI('08-6B-D7-01-DE-81')

        Parsing an incorrect MAC address will fails

        >>> mac_address("08:6b:d7:01:de:666") # doctest:+NORMALIZE_WHITESPACE
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “08:6b:d7:01:de:666” is not a valid MAC
                                    address

    """

    try:
        eui = EUI(address)
    except AddrFormatError as error:
        raise ArgumentTypeError(
            f"“{address}” is not a valid MAC address"
        ) from error

    return eui


def measurement_time(value: str) -> float:
    """Check if the given number is valid measurement time

    0 will be interpreted as infinite measurement runtime

    Returns:

        A float value representing the measurement time on success

    Raises:

        ArgumentTypeError:
             If the given text is not a valid measurement time value


    Examples:

        Parse correct measurement times

        >>> measurement_time("0")
        inf

        >>> measurement_time("0.1")
        0.1

        >>> measurement_time("12.34")
        12.34

        Parsing a negative measurement time fails

        >>> measurement_time("-1")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “-1” is not a valid measurement time

        Parsing an incorrect measurement time fails

        >>> measurement_time("something")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “something” is not a valid measurement time

    """

    try:
        number = float(value)
        if number < 0:
            raise ValueError()
        if number == 0:
            return inf
        return number
    except ValueError as error:
        raise ArgumentTypeError(
            f"“{value}” is not a valid measurement time"
        ) from error


def non_infinite_measurement_time(value: str) -> float:
    """Check if the given number is valid (non-infinite) measurement time

    0 will be interpreted as infinite measurement runtime

    Returns:

        A float value representing the measurement time on success

    Raises:

        ArgumentTypeError:
             If the given text is not a valid measurement time value

    Examples:

        Parse correct measurement time

        >>> non_infinite_measurement_time("0.1")
        0.1

        Parsing an “infinite” measurement time fails

        >>> non_infinite_measurement_time("0") # doctest:+NORMALIZE_WHITESPACE
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “0” is not a valid positive measurement
                                    time

    """

    runtime = measurement_time(value)
    if runtime == inf:
        raise ArgumentTypeError(
            f"“{value}” is not a valid positive measurement time"
        )

    return runtime


def sensor_node_number(value: str) -> int:
    """Check if the given number is valid Bluetooth node number

    Returns:

        An integer representing the given channel number on success

    Raises:

        ArgumentTypeError:
             If the given text is not a valid sensor node number

    Examples:

        Parse correct sensor node numbers

        >>> sensor_node_number("0")
        0

        >>> sensor_node_number("123")
        123

        Parsing an incorrect negative sensor node number fails

        >>> sensor_node_number("-1")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “-1” is not a valid Bluetooth node number

        Parsing an incorrect sensor node number fails

        >>> sensor_node_number("hello") # doctest:+NORMALIZE_WHITESPACE
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “hello” is not a valid Bluetooth node
                                    number

    """

    try:
        number = int(value)
        if number < 0:
            raise ValueError()
        return number
    except ValueError as error:
        raise ArgumentTypeError(
            f"“{value}” is not a valid Bluetooth node number"
        ) from error


def node_name(name: str) -> str:
    """Check if the given text is a valid node name

    Returns:

        The given name on success

    Raises:

        An argument type error in case the given text does not store a valid
        node name. This is the case, if the name

        - is longer than 8 characters or
        - contains non-ASCII data.

    Examples:

        Parse a correct node name

        >>> node_name("Blubb")
        'Blubb'

        A node name is not allowed to contain non-ASCII data

        >>> node_name("Blübb")
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “Blübb” is not a valid name

        A node name can not be longer than 8 characters

        >>> node_name("123456789") # doctest:+NORMALIZE_WHITESPACE
        Traceback (most recent call last):
           ...
        argparse.ArgumentTypeError: “123456789” is too long to be a valid
                                    name

    """

    try:
        name.encode("ascii")
    except UnicodeEncodeError as error:
        raise ArgumentTypeError(f"“{name}” is not a valid name") from error

    if len(name) > 8:
        raise ArgumentTypeError(f"“{name}” is too long to be a valid name")

    return name


def add_identifier_arguments(parser: ArgumentParser) -> None:
    """Add node identifier arguments to given argument parser

    Args:

        parser:
            The parser which should include the node identifier arguments

    """

    identifier_arg_group = parser.add_argument_group(
        title="Sensor Node Identifier"
    )

    identifier_group = identifier_arg_group.add_mutually_exclusive_group(
        required=True
    )

    identifier_group.add_argument(
        "-n",
        "--name",
        dest="identifier",
        metavar="NAME",
        help="Name of sensor node",
        default="Test-STH",
        type=node_name,
    )
    identifier_group.add_argument(
        "-m",
        "--mac-address",
        dest="identifier",
        metavar="MAC_ADRESS",
        help="Bluetooth MAC address of sensor node",
        type=mac_address,
    )
    identifier_group.add_argument(
        "-d",
        "--number",
        type=sensor_node_number,
        dest="identifier",
        metavar="NUMBER",
        help="Node number of sensor node",
    )


def add_adc_arguments(parser: ArgumentParser) -> None:
    """Add ADC arguments to given argument parser

    Args:

        parser:
            The parser which should include the ADC arguments

    """

    adc_group = parser.add_argument_group(title="ADC")
    adc_group.add_argument(
        "-s",
        "--prescaler",
        type=int,
        choices=range(2, 128),
        metavar="2–127",
        default=2,
        required=False,
        help="Prescaler value",
    )
    adc_group.add_argument(
        "-a",
        "--acquisition",
        type=int,
        choices=sorted([2**number for number in range(9)] + [3]),
        default=8,
        required=False,
        help="Acquisition time value",
    )
    adc_group.add_argument(
        "-o",
        "--oversampling",
        type=int,
        choices=[2**number for number in range(13)],
        default=64,
        required=False,
        help="Oversampling rate value",
    )
    adc_group.add_argument(
        "-v",
        "--voltage-reference",
        choices=ADCConfiguration.REFERENCE_VOLTAGES,
        type=float,
        default=3.3,
        required=False,
        help="Reference voltage in V",
    )


def add_channel_arguments(group) -> None:
    """Add channel arguments to given argument parser group

    Args:

        group:
            The parser group to which the channel arguments should be added to

    """

    group.add_argument(
        "-1",
        "--first-channel",
        type=channel_number,
        default=1,
        const=1,
        nargs="?",
        help=(
            "sensor channel number for first measurement channel "
            "(1 - 255; 0 to disable)"
        ),
    )
    group.add_argument(
        "-2",
        "--second-channel",
        type=channel_number,
        default=0,
        const=2,
        nargs="?",
        help=(
            "sensor channel number for second measurement channel "
            "(1 - 255; 0 to disable)"
        ),
    )
    group.add_argument(
        "-3",
        "--third-channel",
        type=channel_number,
        default=0,
        const=3,
        nargs="?",
        help=(
            "sensor channel number for third measurement channel "
            "(1 - 255; 0 to disable)"
        ),
    )


def create_icon_parser() -> ArgumentParser:
    """Create command line parser for icon

    Returns:

        A parser for the CLI arguments of icon

    """

    parser = ArgumentParser(description="ICOtronic CLI tool")

    parser.add_argument(
        "--log",
        choices=("debug", "info", "warning", "error", "critical"),
        default="warning",
        required=False,
        help="minimum log level",
    )

    subparsers = parser.add_subparsers(
        required=True, title="Subcommands", dest="subcommand"
    )

    # ==========
    # = Config =
    # ==========

    subparsers.add_parser(
        "config", help="Open config file in default application"
    )

    # =============
    # = Data Loss =
    # =============

    dataloss_parser = subparsers.add_parser(
        "dataloss", help="Check data loss at different sample rates"
    )
    measurement_group = dataloss_parser.add_argument_group(title="Measurement")
    measurement_group.add_argument(
        "-t",
        "--time",
        type=non_infinite_measurement_time,
        help="Measurement time in seconds",
        default=10,
    )
    add_identifier_arguments(dataloss_parser)

    # ========
    # = List =
    # ========

    subparsers.add_parser("list", help="List sensor nodes")

    # ===========
    # = Measure =
    # ===========

    measurement_parser = subparsers.add_parser(
        "measure", help="Store measurement data"
    )

    measurement_group = measurement_parser.add_argument_group(
        title="Measurement"
    )

    measurement_group.add_argument(
        "-t",
        "--time",
        type=measurement_time,
        help="Measurement time in seconds (0 for infinite runtime)",
        default=10,
    )

    add_channel_arguments(measurement_group)
    add_identifier_arguments(measurement_parser)
    add_adc_arguments(measurement_parser)

    # ==========
    # = Rename =
    # ==========

    rename_parser = subparsers.add_parser(
        "rename", help="Rename a sensor node"
    )
    add_identifier_arguments(rename_parser)
    rename_parser.add_argument(
        "name",
        type=str,
        help="New name of sensor node",
        nargs="?",
        default="Test-STH",
    )

    # =======
    # = STU =
    # =======

    stu_parser = subparsers.add_parser(
        "stu", help="Execute commands related to stationary receiver unit"
    )

    stu_subparsers = stu_parser.add_subparsers(
        required=True, title="Subcommands", dest="stu_subcommand"
    )
    stu_subparsers.add_parser("mac", help="Show Bluetooth MAC address")
    stu_subparsers.add_parser("reset", help="Reset STU")

    return parser


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Listeners used to react to different messages from the CAN bus"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from asyncio import Queue
from collections.abc import Sequence
from logging import getLogger
from typing import NamedTuple

from can import Listener, Message as CANMessage

from icotronic.can.protocol.message import Message
from icotronic.config import settings
from icotronic.utility.log import get_log_file_handler

# -- Classes ------------------------------------------------------------------


class Response(NamedTuple):
    """Used to store a response (message)"""

    message: CANMessage
    """The response message"""

    is_error: bool
    """States if the response was an error or a normal response"""

    error_message: str
    """Optional explanation for the error reason"""


class Logger(Listener):
    """Log ICOtronic CAN messages in a machine and human readable format"""

    def __init__(self):
        self.logger = getLogger("icotronic.can")
        self.logger.propagate = False
        # We use `Logger` in the code below, since the `.logger` attribute
        # stores internal DynaConf data
        self.logger.setLevel(settings.Logger.can.level)
        self.logger.addHandler(get_log_file_handler("can.log"))

    def on_message_received(self, msg: CANMessage) -> None:
        """React to a received message on the bus

        Args:

            msg:
                The received CAN message the notifier should react to

        """

        self.logger.debug("%s", Message(msg))

    def on_error(self, exc: Exception) -> None:
        """Handle any exception in the receive thread.

        Args:

            exc:
                The exception causing the thread to stop

        """

        self.logger.error("Error while monitoring CAN bus data: %s", exc)

    def stop(self) -> None:
        """Stop handling new messages"""


class ResponseListener(Listener):
    """A listener that reacts to messages containing a certain id

    Args:

        message:
            The sent message this listener should react to

        expected_data:
           This optional field specifies the expected acknowledgment data.
           You can either specify to:

           - not check the message data (``None``),
           - check the first bytes by providing a bytearray,
           - check the first bytes by providing a heterogenous list
             of numbers (data byte will be checked for equality) and
             ``None`` (data byte will not be checked).
    """

    def __init__(
        self,
        message: Message,
        expected_data: bytearray | Sequence[int | None] | None,
    ) -> None:

        self.queue: Queue[Response] = Queue()
        identifier = message.identifier()
        self.acknowledgment_identifier = identifier.acknowledge()
        self.error_identifier = identifier.acknowledge(error=True)
        self.expected_data = expected_data

    def on_message_received(self, msg: CANMessage) -> None:
        """React to a received msg on the bus

        Args:

            msg:
                The received CAN message the notifier should react to

        """

        identifier = msg.arbitration_id
        error_response = identifier == self.error_identifier.value
        normal_response = identifier == self.acknowledgment_identifier.value

        # We only store CAN messages that contain the expected (error) response
        # message identifier

        # Also set an error response, if the retrieved message data does not
        # match the expected data
        expected_data = self.expected_data
        error_reason = ""
        if normal_response and expected_data is not None:
            error_response |= any(
                expected != data
                for expected, data in zip(expected_data, msg.data)
                if expected is not None
            )
            error_reason = (
                "Unexpected response message data:\n"
                f"Expected: {list(expected_data)}\n"
                f"Received: {list(msg.data)}"
            )
        elif error_response:
            error_reason = "Received error response"

        if error_response or normal_response:
            self.queue.put_nowait(
                Response(
                    message=msg,
                    is_error=error_response,
                    error_message=error_reason,
                )
            )

    async def on_message(self) -> Response | None:
        """Return answer messages for the specified message identifier

        Returns:

            A response containing

            - the response message for the message with the identifier given at
              object creation, and
            - the error status of the response message

        """

        return await self.queue.get()

    def on_error(self, exc: Exception) -> None:
        """Handle any exception in the receive thread

        Args:

            exc:
                The exception causing the thread to stop

        """

        getLogger().error("Error while monitoring CAN bus data: %s", exc)

    def stop(self) -> None:
        """Stop handling new messages"""


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Communicate with the ICOtronic system"""

# -- Imports ------------------------------------------------------------------

from __future__ import annotations

from asyncio import wait_for
from logging import getLogger

from can import BusABC, Message as CANMessage, Notifier
from netaddr import EUI

from icotronic.can.constants import SENSOR_NODE_NUMBER_SELF_ADDRESSING
from icotronic.can.protocol.message import Message
from icotronic.can.error import ErrorResponseError, NoResponseError
from icotronic.can.listener import ResponseListener
from icotronic.can.node.id import NodeId
from icotronic.test.misc import skip_hardware_tests_ci
from icotronic.utility.data import convert_bytes_to_text

pytestmark = skip_hardware_tests_ci()


# -- Classes ------------------------------------------------------------------


class SPU:
    """Communicate with the ICOtronic system acting as SPU

    Args:

        bus:
            A CAN bus object used to communicate with the STU

        notifier:
            A notifier class that listens to the communication of ``bus``

    """

    def __init__(self, bus: BusABC, notifier: Notifier) -> None:

        self.bus = bus
        self.notifier = notifier
        self.id = NodeId("SPU 1")

    # pylint: disable=too-many-arguments, too-many-positional-arguments

    async def request(
        self,
        message: Message,
        description: str,
        response_data: bytearray | list[int | None] | None = None,
        minimum_timeout: float = 0,
        retries: int = 10,
    ) -> CANMessage:
        """Send a request message and wait for the response

        Args:

            message:
                The message containing the request

            description:
                A description of the request used in error messages

            response_data:
               Specifies the expected data in the acknowledgment message

            minimum_timeout:
               Minimum time before attempting additional connection attempt
               in seconds

            retries:
               The number of times the message is sent again, if no response
               was sent back in a certain amount of time

        Returns:

            The response message for the given request

        Raises:

            NoResponseError
                If the receiver did not respond to the message after retries
                amount of messages sent

            ErrorResponseError
                If the receiver answered with an error message

        """

        logger = getLogger()
        logger.debug("Send request to %s", description)

        for attempt in range(retries):
            listener = ResponseListener(message, response_data)
            self.notifier.add_listener(listener)
            getLogger("icotronic.can").debug("%s", message)
            self.bus.send(message.to_python_can())

            try:
                # We increase the timeout after the first and second try.
                # This way we reduce the chance of the warning:
                #
                # - “Bus error: an error counter reached the 'heavy'/'warning'
                #   limit”
                #
                # happening. This warning might show up after
                #
                # - we flashed the STU,
                # - sent a reset command to the STU, and then
                # - wait for the response of the STU.
                timeout = max(min(attempt * 0.1 + 0.5, 2), minimum_timeout)
                response = await wait_for(
                    listener.on_message(), timeout=timeout
                )
                assert response is not None
            except TimeoutError:
                continue
            finally:
                listener.stop()
                self.notifier.remove_listener(listener)

            if response.is_error:
                raise ErrorResponseError(
                    "Received unexpected response for request to "
                    f"{description}:\n\n{response.error_message}\n"
                    f"Response Message: {Message(response.message)}"
                )

            logger.debug("Retrieved answer for request to %s", description)
            return response.message

        raise NoResponseError(f"Unable to {description}")

    async def request_bluetooth(
        self,
        node: str | NodeId,
        subcommand: int,
        description: str,
        sensor_node_number: int | None = None,
        data: list[int] | None = None,
        response_data: list[int | None] | None = None,
    ) -> CANMessage:
        """Send a request for a certain Bluetooth command

        Args:

            node:
                The node on which the Bluetooth command should be executed

            subcommand:
                The number of the Bluetooth subcommand

            sensor_node_number:
                The node number of the Bluetooth node

            description:
                A description of the request used in error messages

            data:
                An optional list of bytes that should be included in the
                request

            response_data:
                An optional list of expected data bytes in the response message

        Returns:

            The response message for the given request

        """

        sensor_node_number = (
            0 if sensor_node_number is None else sensor_node_number
        )
        data = [0] * 6 if data is None else data
        message = Message(
            block="System",
            block_command="Bluetooth",
            sender=self.id,
            receiver=node,
            request=True,
            data=[subcommand, sensor_node_number] + data,
        )

        # The Bluetooth subcommand and node number should be the same in the
        # response message.
        #
        # Unfortunately the node number is currently not the same for:
        #
        # - the subcommand that sets the second part of the name, and
        # - the subcommand that retrieves the MAC address
        # - the subcommand that writes the time values for the reduced energy
        #   mode
        #
        # The subcommand number in the response message for the commands to
        # set the time values for
        #
        # - the reduced energy mode and
        # - the lowest energy mode
        #
        # are unfortunately also not correct.
        set_second_part_name = 4
        set_times_reduced_energy = 14
        set_times_reduced_lowest = 16
        get_mac_address = 17
        expected_data: list[int | None]
        if subcommand in {get_mac_address, set_second_part_name}:
            expected_data = [subcommand, None]
        elif subcommand in {
            set_times_reduced_energy,
            set_times_reduced_lowest,
        }:
            expected_data = [None, None]
        else:
            expected_data = [subcommand, sensor_node_number]

        if response_data is not None:
            expected_data.extend(response_data)

        return await self.request(
            message, description=description, response_data=expected_data
        )

    async def request_product_data(
        self,
        block_command: str | int,
        description: str,
        node: str | NodeId,
    ) -> CANMessage:
        """Send a request for product data

        Args:

            node:
                The node on which the block command should be executed

            block_command:
                The name or number of the block command

            description:
                A description of the request used in error messages

        Returns:

            The response message for the given request

        """

        message = Message(
            block="Product Data",
            block_command=block_command,
            sender=self.id,
            receiver=node,
            request=True,
            data=[0] * 8,
        )

        return await self.request(message, description=description)

    # pylint: enable=too-many-arguments, too-many-positional-arguments

    async def get_name(
        self,
        node: str | NodeId = "STU 1",
        sensor_node_number: int = SENSOR_NODE_NUMBER_SELF_ADDRESSING,
    ) -> str:
        """Retrieve the name of a Bluetooth node

        You can use this method to name of both

        1. disconnected and
        2. connected

        nodes.

        1. For disconnected sensor nodes you will usually use the STU (e.g.
           ``STU 1``) and the node number at the STU (in the range ``0`` up to
           the number of nodes - 1) to retrieve the name.

        2. For connected nodes you will use the node name and the special
           “self addressing” node number (``0xff``) to ask a node about its
           own name. **Note**: A connected STH will return its own name,
           regardless of the value of the node number.

        Args:

            node:
                The node which has access to the Bluetooth node

            sensor_node_number:
                The number of the Bluetooth node (0 up to the number of
                available nodes - 1; 0xff for self addressing).

        Returns:

            The (Bluetooth broadcast) name of the node

        """

        description = f"name of node “{sensor_node_number}” from “{node}”"

        answer = await self.request_bluetooth(
            node=node,
            subcommand=5,
            sensor_node_number=sensor_node_number,
            description=f"get first part of {description}",
        )

        first_part = convert_bytes_to_text(answer.data[2:])

        answer = await self.request_bluetooth(
            node=node,
            sensor_node_number=sensor_node_number,
            subcommand=6,
            description=f"get second part of {description}",
        )

        second_part = convert_bytes_to_text(answer.data[2:])

        return first_part + second_part

    async def get_rssi(
        self,
        node: str | NodeId = "STH 1",
        sensor_node_number: int = SENSOR_NODE_NUMBER_SELF_ADDRESSING,
    ) -> int:
        """Retrieve the RSSI (Received Signal Strength Indication) of a node

        You can use this method to retrieve the RSSI of both

        1. disconnected and
        2. connected

        nodes.

        1. For disconnected nodes (STHs) you will usually use the STU (e.g.
           ``STU 1``) and the node number at the STU (in the range ``0`` up to
           the number of nodes - 1) to retrieve the RSSI.

        2. For connected nodes you will use the node name and the special
           “self addressing” node number (``0xff``) to ask a node about its
           own RSSI.

        Args:

            node:
                The node which should retrieve the RSSI

            sensor_node_number:
                The number of the Bluetooth node (0 up to the number of
                available nodes - 1; 0xff for self addressing).

        Returns:

            The RSSI of the node specified via node and node number

        Examples:

            Import required library code

            >>> from asyncio import run, sleep
            >>> from icotronic.can.connection import Connection

            Retrieve the RSSI of a disconnected STH

            >>> async def get_bluetooth_rssi():
            ...     async with Connection() as stu:
            ...         await stu.activate_bluetooth()
            ...
            ...         # We assume that at least one STH is available
            ...         # Get the RSSI of node “0”
            ...         return await stu.spu.get_rssi('STU 1', 0)
            >>> rssi = run(get_bluetooth_rssi())
            >>> -70 < rssi < 0
            True

        """

        response = await self.request_bluetooth(
            node=node,
            sensor_node_number=sensor_node_number,
            subcommand=12,
            description=f"get RSSI of “{sensor_node_number}” from “{node}”",
        )

        return int.from_bytes(
            response.data[2:3], byteorder="little", signed=True
        )

    async def get_mac_address(
        self,
        node: str | NodeId = "STH 1",
        sensor_node_number: int = SENSOR_NODE_NUMBER_SELF_ADDRESSING,
    ) -> EUI:
        """Retrieve the Bluetooth MAC address of a node

        You can use this method to retrieve the address of both

        1. disconnected and
        2. connected

        nodes.

        1. For disconnected nodes (STHs) you will usually use the STU (e.g.
           ``STU 1``) and the node number at the STU (in the range ``0`` up to
           the number of nodes - 1) to retrieve the MAC address.

        2. For connected nodes you will use the node name and the special
           “self addressing” node number (``0xff``) to ask a node about its
           own node number. **Note**: A connected STH will return its own
           MAC address, regardless of the value of the node number.

        Args:

            node:
                The node which should retrieve the MAC address

            sensor_node_number:
                The number of the Bluetooth node (0 up to the number of
                available nodes - 1; 0xff for self addressing).

        Returns:

            The MAC address of the node specified via node and node number

        """

        response = await self.request_bluetooth(
            node=node,
            sensor_node_number=sensor_node_number,
            subcommand=17,
            description=(
                f"get MAC address of “{sensor_node_number}” from “{node}”"
            ),
        )

        return EUI(":".join(f"{byte:02x}" for byte in response.data[:1:-1]))


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()



"""Read and store some acceleration data of STH with node name Test-STH"""

# -- Imports ------------------------------------------------------------------

from asyncio import run
from pathlib import Path
from time import monotonic

from netaddr import EUI

from icotronic.can import Connection, StreamingConfiguration, STH
from icotronic.measurement import Storage

# -- Functions ----------------------------------------------------------------


async def store_streaming_data(identifier: EUI | str | int) -> None:
    """Store streaming data in HDF5 file

    Args:

        identifier:

            Identifier of STH node

    """

    async with Connection() as stu:
        async with stu.connect_sensor_node(identifier, STH) as sth:

            assert isinstance(sth, STH)  # Make type checker happy

            conversion_to_g = await sth.get_acceleration_conversion_function()

            filepath = Path("test.hdf5")
            stream_first = StreamingConfiguration(first=True)

            with Storage(filepath, channels=stream_first) as storage:
                # Store acceleration range as metadata
                storage.write_sensor_range(
                    await sth.get_acceleration_sensor_range_in_g()
                )
                # Store sampling rate (and ADC configuration as metadata)
                storage.write_sample_rate(await sth.get_adc_configuration())
                async with sth.open_data_stream(stream_first) as stream:
                    # Read data for about five seconds
                    end = monotonic() + 5
                    async for data, _ in stream:
                        # Convert from ADC bit value into multiples of g
                        storage.add_streaming_data(data.apply(conversion_to_g))
                        if monotonic() > end:
                            break


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    run(store_streaming_data(identifier="Test-STH"))



"""Streaming errors"""

# -- Classes ------------------------------------------------------------------


class StreamingError(Exception):
    """General exception for streaming errors"""


class StreamingTimeoutError(StreamingError):
    """Raised if no streaming data was received for a certain amount of time"""


class StreamingBufferError(StreamingError):
    """Raised if there are too many streaming messages in the buffer"""



"""Data formats for streaming data"""

# -- Imports ------------------------------------------------------------------

from icotronic.can.streaming.config import StreamingConfiguration

# -- Classes ------------------------------------------------------------------


class StreamingFormat:
    """Support for specifying the `data streaming format`_

    .. _data streaming format:
       https://mytoolit.github.io/Documentation/#block-streaming

    Args:

        *value:
            The value of the streaming format byte

        streaming:
            Specifies if this is a request for a stream of data bytes;
            If this value is not set or set to ``False``, then the request is
            only for a single value (or set of values).

        width:
            Specifies the width of a single value (either 2 or 3 bytes)

        channels:
            Specifies for which channels data should be transmitted or not

        sets:
            Specifies the number of data sets that should be transmitted

            The value 0 stops the stream. Other possible values for the
            number of sets are 1, 3, 6, 10, 15, 20 and 30.

        value_explanations:
            Three strings used to describe the first, second and third data
            value

    """

    data_set = [0, 1, 3, 6, 10, 15, 20, 30]
    """Possible number of data sets"""

    # pylint: disable=too-many-arguments

    def __init__(
        self,
        *value,
        streaming: bool | None = None,
        width: int | None = 2,
        channels: StreamingConfiguration | None = None,
        sets: int | None = None,
        value_explanations: tuple[str, str, str] = (
            "Value 1",
            "Value 2",
            "Value 3",
        ),
    ) -> None:

        def set_part(start, width, number):
            """Store bit pattern number at bit start of the identifier"""

            streaming_ones = 0xFF
            mask = (1 << width) - 1

            # Set all bits for targeted part to 0
            self.value &= (mask << start) ^ streaming_ones
            # Make sure we use the correct number of bits for number
            number = number & mask
            # Set bits to given value
            self.value |= number << start

        self.value_explanations = value_explanations

        if len(value) > 1:
            raise ValueError("More than one positional argument")

        self.value = value[0] if value else 0

        # =============
        # = Streaming =
        # =============

        if streaming:
            set_part(7, 1, int(streaming))

        # =========
        # = Width =
        # =========

        if width is not None:
            if not (isinstance(width, int) and 2 <= width <= 3):
                raise ValueError(f"Unsupported width value: {width}")

            set_part(6, 1, 1 if width == 3 else 0)

        # =================
        # = Active Values =
        # =================

        if channels:
            for shift, part in enumerate(
                [channels.third, channels.second, channels.first]
            ):
                if part is not False:
                    set_part(3 + shift, 1, part)

        # =============
        # = Data Sets =
        # =============

        if sets is not None:
            cls = type(self)

            if sets not in cls.data_set:
                raise ValueError(f"Unsupported number of data sets: {sets}")

            set_part(0, 3, cls.data_set.index(sets))

    # pylint: enable=too-many-arguments

    def __repr__(self) -> str:
        """Retrieve the textual representation of the streaming format

        Returns:

            A string that describes the streaming format

        Examples:

            Get the textual representation of some example streaming formats

            >>> StreamingFormat(width=3,
            ...                 channels=StreamingConfiguration(first=True),
            ...                 sets=15)
            Single Request, 3 Bytes, 15 Data Sets, Read Value 1

            >>> StreamingFormat(0b001, streaming=True)
            Streaming, 2 Bytes, 1 Data Set

            >>> StreamingFormat(0b110111)
            Single Request, 2 Bytes, 30 Data Sets, Read Value 1, Read Value 2

        """

        streaming = self.value >> 7

        data_sets = self.data_sets()
        data_set_explanation = (
            "Stop Stream"
            if data_sets == 0
            else f"{data_sets} Data Set{'' if data_sets == 1 else 's'}"
        )

        parts = [
            "Streaming" if streaming else "Single Request",
            f"{self.data_bytes()} Bytes",
            f"{data_set_explanation}",
        ]

        value_selection = (self.value >> 3) & 0b111

        first = value_selection >> 2
        second = value_selection >> 1 & 1
        third = value_selection & 1

        selected_values = [
            f"Read {value_explanation}"
            for selected, value_explanation in zip(
                (first, second, third), self.value_explanations
            )
            if selected
        ]

        value_explanation = (
            ", ".join(selected_values) if selected_values else ""
        )
        if value_explanation:
            parts.append(value_explanation)

        return ", ".join(parts)

    def data_sets(self) -> int:
        """Get the number of data sets of the streaming format

        Returns:

            The number of data sets

        Examples:

            Retrieve the number of data sets for some example streaming formats

            >>> StreamingFormat(
            ...     width=3,
            ...     channels=StreamingConfiguration(first=True),
            ...     sets=15
            ... ).data_sets()
            15

            >>> StreamingFormat(
            ...     channels=StreamingConfiguration(first=True, second=False),
            ...     sets=3
            ... ).data_sets()
            3

        """

        data_set_bits = self.value & 0b111
        cls = type(self)

        return cls.data_set[data_set_bits]

    def data_bytes(self) -> int:
        """Get the number of data bytes used for a single value

        Returns:

            The number of data bytes that represent a single streaming value

        Examples:

            Retrieve the number of data bytes for some streaming formats

            >>> StreamingFormat(width=3,
            ...                 channels=StreamingConfiguration(first=True),
            ...                 sets=15).data_bytes()
            3

            >>> StreamingFormat(
            ...     channels=StreamingConfiguration(first=True, second=False),
            ...     width=2
            ... ).data_bytes()
            2

        """

        return 3 if (self.value >> 6) & 1 else 2


class StreamingFormatVoltage(StreamingFormat):
    """Support for specifying the streaming format of voltage data

    Args:

        value:
            The value of the streaming format byte

        single:
            Specifies if the request was for a single value or not

        width:
            Specifies the width of a single value (either 2 or 3 bytes)

        channels:
            Specifies for which channels data should be transmitted or not

        sets:
            Specifies the number of data sets that should be transmitted

            The value 0 stops the stream. Other possible values for the
            number of sets are 1, 3, 6, 10, 15, 20 and 30.

    """

    def __init__(self, *arguments, **keyword_arguments) -> None:

        super().__init__(
            *arguments,
            **keyword_arguments,
            value_explanations=("Voltage 1", "Voltage 2", "Voltage 3"),
        )



"""Misc functions related to testing"""

# -- Imports ------------------------------------------------------------------

from pytest import MarkDecorator

# -- Functions ----------------------------------------------------------------

# pylint: disable=import-outside-toplevel


def skip_hardware_tests_ci() -> MarkDecorator | None:
    """Skip hardware dependent test on CI sytem

    Returns:

        - ``None`` if pytest is not installed or
        - a decorator that skips tests if the environment variable ``CI`` is
          defined

    """

    try:
        from pytest import mark
        from os import environ

        return mark.skipif(
            "CI" in environ, reason="requires ICOtronic hardware"
        )
    except ModuleNotFoundError:
        return None


# pylint: enable=import-outside-toplevel



"""Code for measuring performance"""

# -- Imports ------------------------------------------------------------------

from time import perf_counter_ns, process_time_ns

# -- Classes ------------------------------------------------------------------


class PerformanceMeasurement:
    """Store information about performance measurements

    Examples:

        Import required library code

        >>> from time import sleep

        Determine performance data for sleep call

        >>> perf_measurement = PerformanceMeasurement()
        >>> perf_measurement.start()
        >>> sleep(0.5)
        >>> perf_measurement.stop()

        >>> 0.5 <= perf_measurement.run_time() < 0.7 * 10**9
        True
        >>> 0 <= perf_measurement.cpu_time() < 0.7 * 10**9
        True
        >>> perf_measurement.cpu_time() <= perf_measurement.run_time()
        True
        >>> 0 <= perf_measurement.cpu_usage() <= 1
        True

    """

    def __init__(self) -> None:
        self.perf_counter_start = 0
        self.process_time_start = 0

        self.perf_counter_end = 0
        self.process_time_end = 0

    def start(self) -> None:
        """Start performance measurement"""

        self.perf_counter_start = perf_counter_ns()
        self.process_time_start = process_time_ns()

        self.perf_counter_end = self.perf_counter_start
        self.process_time_end = self.process_time_start

    def stop(self) -> None:
        """Stop performance measurement"""
        self.perf_counter_end = perf_counter_ns()
        self.process_time_end = process_time_ns()

    def run_time(self) -> int:
        """Determine run time

        Returns:

            The amount of ns that the measured code executed in wall time

        """
        return self.perf_counter_end - self.perf_counter_start

    def cpu_time(self) -> int:
        """Determine CPU time

        Returns:

            The amount of ns that the measured code executed in the CPU

        """

        return self.process_time_end - self.process_time_start

    def cpu_usage(self) -> float:
        """Determine the CPU usage

        Returns:
            The CPU usage as factor between 0 and 1

        """

        return self.cpu_time() / self.run_time()

    def __repr__(self) -> str:
        """Print the textual representation of a performance measurement

        Returns:

            Textual data about performance measurement

        Examples:

            Import required library code

            >>> from time import sleep

            Print performance measurement data

            >>> perf_measurement = PerformanceMeasurement()
            >>> perf_measurement.start()
            >>> sleep(1)
            >>> perf_measurement.stop()
            >>> print(perf_measurement
            ...      ) # doctest:+ELLIPSIS, +NORMALIZE_WHITESPACE
            🏁 Run Time: ... seconds, 💼 CPU time: ... seconds,
            💪 CPU Usage: ... %

        """

        return ", ".join([
            f"🏁 Run Time: {self.run_time() / 10**9:.2f} seconds",
            f"💼 CPU time: {self.cpu_time() / 10**9:.2f} seconds",
            f"💪 CPU Usage: {self.cpu_usage() * 100:.2f} %",
        ])


# -- Main ---------------------------------------------------------------------

if __name__ == "__main__":
    from doctest import testmod

    testmod()
